
import time
import pathlib
import logging
import json

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt, cm, patches, ticker
from matplotlib.colors import LinearSegmentedColormap
import pysam

from triplix.core.header import TriplixHeader
from triplix.core import configurations
from triplix.core import concatemers
from triplix.core.triplets import TripletsContainer
from triplix.core.tri_alignments import TriAlignmentsContainer
from triplix.core.utilities import overlap, triplet_to_cube

COMMAND_NAME = 'triplix.plot'
logger = logging.getLogger(COMMAND_NAME)


def plot_virtual_hic(
        input_patterns,
        view_chrom, view_start, view_end,
        view_point,
        anchor_width=None,
        plot_names=('count_ABC',),
        mapping_quality=5,
        output_dir=None,
        output_name=None,
        plot_params=None,
):
    # prepare plot parameters
    parameters = {}
    for plot_name in plot_names:
        if plot_name.endswith('.enrichment'):
            parameters[plot_name] = {
                'clim_range': [-4, 4],
                'draw_values': True,
            }
        else:
            parameters[plot_name] = {
                'clim_prct': [20, 96],
                'draw_values': True,
            }
    for plot_name in plot_params:
        for key, value in plot_params[plot_name].items():
            parameters[plot_name][key] = value

    # casting input parameters to proper type casting
    input_patterns = input_patterns.split(',')
    view_start = int(view_start)
    view_end = int(view_end)
    view_point = int(view_point)

    # collecting input files
    logger.debug(f'Collecting input files from {len(input_patterns)} given patterns:')
    input_files = []
    for pi, pattern in enumerate(input_patterns):
        logger.debug(f'Pattern #{pi + 1}: {pattern}')
        path = pathlib.Path(pattern).expanduser()
        input_files.extend(path.parent.glob(path.name))
    logger.debug(f'Found {len(input_files):d} input files:')
    for file_idx, file_path in enumerate(input_files):
        logger.debug(f'\t{file_idx + 1:2d}: {file_path}')
    assert len(input_files) > 0, 'At least a single input file is required for plotting'

    # define output directory and filename
    if output_dir is None:
        output_dir = input_files[0].parent
    if output_name is None:
        output_name = (
            f'virtual-hic.{view_chrom}'
            f'_{view_start:09,d}-{view_end:09,d}'
            f'.vp-{view_point:09,d}'
            f'.{input_files[0].name}.{plot_names[-1]}'
            f'.nf{len(input_files):d}.np{len(plot_names):d}'
            f'.pdf')
    output_path = pathlib.Path(output_dir).expanduser() / output_name
    output_path.parent.mkdir(parents=True, exist_ok=True)

    # collecting metadata
    chrom_lengths = None
    for input_idx, input_path in enumerate(input_files):

        # collecting the header and anchor width if needed
        input_header = TriplixHeader(file_path=str(input_path))
        if input_header.anchor_width is not None:
            if anchor_width is None:
                anchor_width = input_header.anchor_width
            else:
                assert anchor_width == input_header.anchor_width, 'At least one file has an incompatible `anchor_width`'
        if input_header.chrom_lengths is not None:
            if chrom_lengths is None:
                chrom_lengths = input_header.chrom_lengths
            else:
                assert chrom_lengths == input_header.chrom_lengths, 'At least one file has an incompatible `chrom_length`'
    assert anchor_width is not None, '`anchor_width` should be defined, either in CLI arguments, or input files.'
    assert chrom_lengths is not None, '`chrom_lengths` should be defined in at least one input file.'
    anchor_width = int(anchor_width)

    # define the chunk area
    if view_start > chrom_lengths[view_chrom]:
        logger.info(f'The requested window ({view_start / 1e6:0.2f}-{view_end / 1e6:0.2f}mb) '
                    f'is outside of {view_chrom} (length={chrom_lengths[view_chrom] / 1e6:0.2f}mb). '
                    f'No processing is needed.')
        return
    assert view_start <= view_point < view_end, 'The viewpoint is out of the view area'
    logger.info(f'Area of interest is: '
                f'{view_chrom}:{view_start:0,d}-{view_end:0,d} | vp={view_point:0,d}. ')

    # define anchors
    anchor_edges = np.arange(view_start, view_end + anchor_width, anchor_width, dtype=int)
    vp_bdx = np.searchsorted(anchor_edges, view_point, side='right') - 1
    n_anchor = len(anchor_edges) - 1

    # collecting counts from each input file
    images = {col_name: np.full([n_anchor, n_anchor], fill_value=np.nan) for col_name in plot_names}
    for input_idx, input_path in enumerate(input_files):
        logger.info(f'{input_idx + 1:d}/{len(input_files):d}: Processing: {input_path}')

        # if the input is a BAM file
        if input_path.suffix in ['.bam']:

            # check file to be sorted and indexed
            if not pathlib.Path(str(input_path) + '.bai').is_file():
                raise FileNotFoundError(f'Missing index file. Please index the coord-sorted BAM file using `samtools index` command: {input_path}')
            bam_header = TriplixHeader(file_path=str(input_path))
            if bam_header.process_history.headers['HD']['SO'] != 'coordinate':
                raise ValueError(f'BAM files need to be "coordinate" sorted. Please sort using `samtools sort` command: {input_path}')

            # collect read ids
            logger.debug('\tCollecting "read_names" that contain the viewpoint ...')
            read_ids = set()
            with pysam.AlignmentFile(input_path, 'r', require_index=True) as bam_fid:
                alignment_iter = bam_fid.fetch(view_chrom, anchor_edges[vp_bdx], anchor_edges[vp_bdx + 1])  # start and stop denote 0-based, half-open intervals, i.e., [start, end).
                for aln_idx, aln in enumerate(alignment_iter):
                    if aln.mapping_quality >= mapping_quality:
                        read_ids.add(aln.query_name)
            logger.debug(f'\t{len(read_ids):,d} reads contain the viewpoint.')
            if len(read_ids) == 0:
                logger.warning(f'No read contains the viewpoint, ignoring: {input_path}')
                continue

            # collecting interactions
            logger.debug('\tCollecting interactions of the selected reads ...')
            alignments = []
            n_ignore = 0
            logger.log_time = time.time()
            with pysam.AlignmentFile(input_path, 'r', require_index=True) as bam_fid:
                alignment_iter = bam_fid.fetch(view_chrom, view_start, view_end)
                n_load = 0
                for aln_idx, aln in enumerate(alignment_iter):
                    if time.time() - logger.log_time > configurations.configs['log_interval']:
                        logger.debug(f'\t\tcurrently processing {aln_idx:d}th alignment')
                        logger.log_time = time.time()
                    if aln.query_name not in read_ids:
                        continue
                    if aln.mapping_quality < mapping_quality:
                        n_ignore += 1
                        continue
                    n_load += 1

                    alignment = dict(
                        id=aln.query_name,
                        chrom=aln.reference_name,
                        start=aln.reference_start,
                        end=aln.reference_end,
                        mapq=aln.mapping_quality,
                    )
                    alignments.append(alignment)
            logger.debug(f'\tIn total {n_load:d} alignments are collected, and {n_ignore:,d} are ignored.')

            # conversion to DataFrame
            alignments = pd.DataFrame(alignments)
            assert alignments['id'].isin(read_ids).all()
            assert np.isin(list(read_ids), alignments['id']).all()

            # identify the VP alignments
            # alignments['is_vp'] = False
            # is_vp = (
            #         (alignments['start'] < anchor_edges[vp_bdx + 1])
            #         & (alignments['end'] >= anchor_edges[vp_bdx])
            # )
            # assert np.array_equal(sorted(set(alignments.loc[is_vp, 'id'])),
            #                       sorted(set(alignments['id'])))
            # alignments.loc[is_vp, 'is_vp'] = True
            # alignments = alignments.sort_values(by=['id', 'is_vp', 'start'], ascending=[True, False, True]).reset_index(drop=True)
            # alignments = alignments[~is_vp].reset_index(drop=True)
            # logger.debug(f'\t{len(alignments):,d} alignments are left after viewpoint exclusion.')

            # sort, and map alignments to bins
            alignments = alignments.sort_values(by=['id', 'start']).reset_index(drop=True)
            alignments['start_bidx'] = np.searchsorted(anchor_edges, alignments['start'], side='right') - 1
            alignments['end_bidx'] = np.searchsorted(anchor_edges, alignments['end'], side='right') - 1

            # add alignments to the counts
            logger.debug('\tMapping alignments to bins')
            n_dup = 0
            seen_ids = dict()
            n_contact = 0
            alignment_grp = alignments.groupby(by='id', sort=False)
            for read_idx, (read_id, read) in enumerate(alignment_grp):
                n_frag = len(read)
                if n_frag < 3:
                    continue

                # store tri-alignments (ignoring the first viewpoint alignment)
                # assert read['is_vp'].iat[0]
                for fi in range(n_frag):
                    frg_i = read.iloc[fi]
                    for fj in range(fi + 1, n_frag):
                        frg_j = read.iloc[fj]
                        for fk in range(fj + 1, n_frag):
                            frg_k = read.iloc[fk]
                            # if np.abs(frg_j['start'] - frg_i['start']) < cli_args.min_dist:
                            #     continue
                            for ovl_i in range(frg_i['start_bidx'], frg_i['end_bidx'] + 1):
                                if not 0 <= ovl_i < n_anchor:
                                    continue
                                for ovl_j in range(frg_j['start_bidx'], frg_j['end_bidx'] + 1):
                                    if not 0 <= ovl_j < n_anchor:
                                        continue
                                    for ovl_k in range(frg_k['start_bidx'], frg_k['end_bidx'] + 1):
                                        if not 0 <= ovl_k < n_anchor:
                                            continue

                                        # assigning a pair id
                                        pair_id = list(sorted([ovl_i, ovl_j, ovl_k]))
                                        if not any(i == vp_bdx for i in pair_id):
                                            continue
                                        pair_id.pop(pair_id.index(vp_bdx))
                                        pair_id = tuple(pair_id)

                                        # ignore if the read has contributed to the current bin-triplet already
                                        if pair_id not in seen_ids:
                                            seen_ids[pair_id] = set()
                                        if read_id in seen_ids[pair_id]:
                                            n_dup += 1
                                            continue
                                        seen_ids[pair_id].add(read_id)

                                        # add the counts
                                        bdx_j, bdx_k = pair_id
                                        if np.isnan(images['count_ABC'][bdx_j, bdx_k]):
                                            images['count_ABC'][bdx_j, bdx_k] = 0
                                            images['count_ABC'][bdx_k, bdx_j] = 0
                                        images['count_ABC'][bdx_j, bdx_k] += 1
                                        if bdx_j != bdx_k:
                                            images['count_ABC'][bdx_k, bdx_j] += 1
                                        n_contact += 1

                                    # ###
                                    # if ovl_i == 1 and ovl_j == 20:
                                    #     print(read[['id', 'start', 'end', 'start_bidx', 'end_bidx']])
                                    #     print(images['count_ABC'][ovl_i, ovl_j])

            # seen_ids_collection[bam_fpath] = copy.deepcopy(seen_ids)
            logger.info(f'\t{n_contact:,d} contacts are stored, and {n_dup:,d} are ignored due to duplicity.')

        # tri-alignment input
        elif input_path.suffix in ['.tri-alignments.tsv.bgz', '.btrp']:
            logger.info(f'Loading tri-alignments from: {input_path}')
            tri_alignments_obj = TriAlignmentsContainer(trialn_path=input_path)

            # map tri-alignments to data cubes
            anchor_max_distance = anchor_edges[-1] - anchor_edges[0]
            counts = np.zeros([n_anchor, n_anchor, n_anchor], dtype=int)
            n_load = 0
            logger.log_time = 0
            for ai in range(n_anchor):
                if time.time() - logger.log_time > configurations.configs['log_interval']:
                    logger.log_time = time.time()
                    logger.info(f'\tcurrently at {anchor_edges[ai]:,d} ({ai:3d}/{n_anchor}) anchor, and loaded {n_load:,d} (unique) tri-alignments')
                counts[ai, :, :] = tri_alignments_obj.fetch_slice(
                    vp_chrom=view_chrom, vp_start=anchor_edges[ai], vp_end=anchor_edges[ai + 1] - 1,
                    anchor_edges=anchor_edges,
                    anchor_max_distance=anchor_max_distance,
                    mapping_quality=mapping_quality
                )
                n_load += np.sum(counts[ai, :, :])

            # make the counts symmetric
            logger.debug('Making the counts cube symmetric')
            for ai in range(n_anchor):
                for aj in range(ai, n_anchor):
                    for ak in range(aj, n_anchor):
                        counts[ai, ak, aj] = counts[ai, aj, ak]
                        counts[aj, ai, ak] = counts[ai, aj, ak]
                        counts[aj, ak, ai] = counts[ai, aj, ak]
                        counts[ak, ai, aj] = counts[ai, aj, ak]
                        counts[ak, aj, ai] = counts[ai, aj, ak]

            # calculate triplet properties
            cubes = {}
            for col_name in plot_names:
                cubes[col_name] = np.full([n_anchor, n_anchor, n_anchor], fill_value=np.nan)
                if col_name == 'count_ABC':
                    cubes[col_name] = counts.copy()
                else:
                    logger.warning(f'Estimation of {col_name} from tri-alignments is not implemented. This map will be all NaNs.')

            # collect data from the viewpoint slice
            for col_name in plot_names:
                has_value = np.isnan(images[col_name]) & ~ np.isnan(cubes[col_name][vp_bdx])
                images[col_name][has_value] = 0
                images[col_name][has_value] += cubes[col_name][vp_bdx][has_value]

        elif str(input_path).endswith('.triplets.h5'):
            logger.info(f'Loading triplets from: {input_path}')

            # prepare Triplets object
            triplets_obj = TripletsContainer(input_path)
            triplets = triplets_obj.fetch(
                chrom_a=view_chrom, start_a=anchor_edges[0], end_a=anchor_edges[-1] - 1,
                chrom_b=view_chrom, start_b=anchor_edges[0], end_b=anchor_edges[-1] - 1,
                chrom_c=view_chrom, start_c=anchor_edges[0], end_c=anchor_edges[-1] - 1,
                exclusive=True
            )

            # map triplets to data slices
            slice_maps = {}
            for col_name in plot_names:
                slice_maps[col_name] = triplets.to_slice(view_point=view_point, anchor_edges=anchor_edges, value_column=col_name)

            # collect data from the viewpoint slice
            for col_name in plot_names:
                has_value = np.isnan(images[col_name]) & ~ np.isnan(slice_maps[col_name])
                images[col_name][has_value] = 0
                images[col_name][has_value] += slice_maps[col_name][has_value]
        else:
            raise ValueError(f'Unknown file type: {input_path}')
    # n_load = np.nansum(image)
    # assert n_load > 0, 'No contacts were found in this area.'

    # filter NaN values
    if 'count_ABC' in images:
        is_nan = np.isnan(images['count_ABC'])
        images['count_ABC'][is_nan] = 0

    # prepare a figure
    ext_bnd = [view_start, view_end]
    # n_row = int(np.ceil(len(plot_names) / 4))
    # n_col = min(len(plot_names), 4)
    n_row = 1
    n_col = len(plot_names)
    plt.close('all')
    fig_h, axes = plt.subplots(nrows=n_row, ncols=n_col, figsize=(n_col * 10, n_row * 8 + 1))
    if hasattr(axes, '__len__'):
        axes = axes.flatten()
    else:
        axes = [axes]
    for pi, col_name in enumerate(plot_names):

        # prepare a colormap
        if col_name.endswith('.enrichment'):
            axes[pi].cmap = LinearSegmentedColormap.from_list('bwwr', ['#0000ff', '#ffffff', '#ff0000'])
            axes[pi].cmap.set_over(color='#700000', alpha=1.0)
            axes[pi].cmap.set_under(color='#000070', alpha=1.0)
            axes[pi].clim = parameters[col_name]['clim_range']
        else:
            axes[pi].cmap = cm.get_cmap('hot_r', 20)
            axes[pi].cmap.set_under(color='#ffffff', alpha=1.0)
            # cmap.set_over(color='#000055', alpha=1.0)
        axes[pi].cmap.set_bad(color='#b2b2a3', alpha=0.5)

        # draw the image
        axes[pi].img_h = axes[pi].imshow(
            images[col_name],
            interpolation='nearest',
            cmap=axes[pi].cmap, aspect='auto',
            extent=ext_bnd + ext_bnd[::-1],
        )

        axes[pi].cb_h = plt.colorbar(axes[pi].img_h, ax=axes[pi], fraction=0.025, pad=0.03, extend='both')
        # axes[pi].cb_h.ax.tick_params(labelsize=10)

        # set color ranges
        if 'clim_range' in parameters[col_name]:
            axes[pi].clim = parameters[col_name]['clim_range']
        elif 'clim_prct' in parameters[col_name]:
            prct_range = parameters[col_name]['clim_prct']
            axes[pi].clim = np.nanpercentile(images[col_name], prct_range)
        else:
            raise ValueError('Can not determine the `clim` variable range.')
        axes[pi].img_h.set_clim(axes[pi].clim)

        # add viewpoint bars
        axes[pi].add_patch(patches.Rectangle((ext_bnd[0], view_point), ext_bnd[1] - ext_bnd[0], anchor_width,
                                             linewidth=0, edgecolor='None', facecolor='#66a96c', alpha=1.0))
        axes[pi].add_patch(patches.Rectangle((view_point, ext_bnd[0]), anchor_width, ext_bnd[1] - ext_bnd[0],
                                             linewidth=0, edgecolor='None', facecolor='#66a96c', alpha=1.0))

        x_tick_pos = np.unique(np.linspace(view_start, view_end, 25, dtype=int))
        axes[pi].set_xticks(x_tick_pos)
        axes[pi].set_yticks(x_tick_pos)
        axes[pi].xaxis.set_tick_params(rotation=60)
        # axes[pi].xaxis.set_major_locator(ticker.MaxNLocator(nbins=17, integer=False))
        # axes[pi].yaxis.set_major_locator(ticker.MaxNLocator(nbins=17, integer=False))
        axes[pi].xaxis.set_major_formatter(lambda v, _: '{:0.2f}'.format(v / 1e6))
        axes[pi].yaxis.set_major_formatter(lambda v, _: '{:0.2f}'.format(v / 1e6))
        # axes[pi].xaxis.set_tick_params(labelsize=10)
        # axes[pi].yaxis.set_tick_params(labelsize=10)

        axes[pi].set_title(f'{col_name}\nsum={np.nansum(images[col_name]):0.1f}')

        # drawing the values over map
        if parameters[col_name]['draw_values']:
            norm = plt.Normalize(*axes[pi].clim)
            for aj in range(n_anchor):
                for ak in range(n_anchor):
                    value = images[col_name][aj, ak]
                    if np.round(value, decimals=1) == 0:
                        continue

                    # define the color from luminance
                    # source: https://stackoverflow.com/questions/596216/formula-to-determine-perceived-brightness-of-rgb-color
                    color_rgb = axes[pi].cmap(norm(value))
                    luminance = color_rgb[0] * 0.2126 + color_rgb[1] * 0.7152 + color_rgb[2] * 0.0722
                    if luminance < 0.549019:  # 0.549019 = 140 / 255
                        text_color = '#FFFFFF'
                    else:
                        text_color = '#000000'

                    # plotting the value
                    axes[pi].text(
                        anchor_edges[ak] + anchor_width // 2,
                        anchor_edges[aj] + anchor_width // 2,
                        f'{value:0.1f}', color=text_color, fontsize=1, zorder=100, va='center', ha='center',
                    )

    plt.suptitle(f'{input_files[0].name}, #files used: {len(input_files)}\n'
                 f'{view_chrom}:{view_start / 1e6:0,.3f}-{view_end / 1e6:0,.3f}mb, viewpoint={view_point / 1e6:0,.3f}mb\n'
                 f'anchor width={anchor_width / 1e3:0.0f}k; mapping quality={mapping_quality}',
                 fontweight='normal')  # y=1 + 0.15 / (n_row * 3)
    plt.subplots_adjust(top=1.0 - 0.17 / n_row, wspace=0.2, hspace=0.2)  # , top=0.8
    # plt.show()

    # logger_fonttools = logging.getLogger('fontTools.subset')
    # logger_fonttools = logging.getLogger('matplotlib')
    # loglevel_prev = logger_fonttools.level
    # logger_fonttools.setLevel(logging.WARNING)
    plt.savefig(output_path, bbox_inches='tight')
    # logger_fonttools.setLevel(loglevel_prev)
    plt.close()
    logger.info(f'Virtual-HiC plot is stored in: {output_path}')


def plot_hic(
        input_patterns,
        view_chrom, view_start, view_end,
        anchor_width=None,
        plot_names=('count_AB',),
        mapping_quality=5,
        output_dir=None,
        output_name=None,
        plot_params=None,
):
    # casting input parameters to proper type casting
    if plot_params is None:
        plot_params = {}
    input_patterns = input_patterns.split(',')
    view_start = int(view_start)
    view_end = int(view_end)

    # collecting input files
    logger.debug(f'Collecting input files from {len(input_patterns)} given patterns:')
    input_files = []
    for pi, pattern in enumerate(input_patterns):
        logger.debug(f'Pattern #{pi + 1}: {pattern}')
        path = pathlib.Path(pattern).expanduser()
        input_files.extend(path.parent.glob(path.name))
    logger.debug(f'Found {len(input_files):d} input files:')
    for file_idx, file_path in enumerate(input_files):
        logger.debug(f'\t{file_idx + 1:2d}: {file_path}')
    assert len(input_files) > 0, 'At least a single input file is required for plotting'

    # define output directory and filename
    if output_dir is None:
        output_dir = input_files[0].parent
    if output_name is None:
        output_name = f'hic.{view_chrom}_{view_start:09,d}-{view_end:09,d}.{input_files[0].stem}.pdf'
    output_path = pathlib.Path(output_dir).expanduser() / output_name
    output_path.parent.mkdir(parents=True, exist_ok=True)

    # collecting metadata
    chrom_lengths = None
    for input_idx, input_path in enumerate(input_files):

        # collecting the header and anchor width if needed
        input_header = TriplixHeader(file_path=str(input_path))
        if input_header.anchor_width is not None:
            if anchor_width is None:
                anchor_width = input_header.anchor_width
            else:
                assert anchor_width == input_header.anchor_width, 'At least one file has an incompatible `anchor_width`'
        if input_header.chrom_lengths is not None:
            if chrom_lengths is None:
                chrom_lengths = input_header.chrom_lengths
            else:
                assert chrom_lengths == input_header.chrom_lengths, 'At least one file has an incompatible `chrom_length`'
    assert anchor_width is not None, '`anchor_width` should be defined, either in CLI arguments, or input files.'
    assert chrom_lengths is not None, '`chrom_lengths` should be defined in at least one input file.'
    anchor_width = int(anchor_width)

    # define the chunk area
    if view_start > chrom_lengths[view_chrom]:
        logger.info(f'The requested window ({view_start / 1e6:0.2f}-{view_end / 1e6:0.2f}mb) '
                    f'is outside of {view_chrom} (length={chrom_lengths[view_chrom] / 1e6:0.2f}mb). '
                    f'No processing is needed.')
        return
    logger.info(f'Area of interest is: {view_chrom}:{view_start:0,d}-{view_end:0,d}')

    # define anchors
    anchor_edges = np.arange(view_start, view_end + anchor_width, anchor_width, dtype=int)
    n_anchor = len(anchor_edges) - 1

    # collecting counts from each input file
    images = {col_name: np.full([n_anchor, n_anchor], fill_value=np.nan) for col_name in plot_names}
    for input_idx, input_path in enumerate(input_files):
        logger.info(f'{input_idx + 1:d}/{len(input_files):d}: Processing: {input_path}')

        if input_path.suffix in ['.bam']:

            # check file to be sorted and indexed
            if not pathlib.Path(str(input_path) + '.bai').is_file():
                raise FileNotFoundError(f'Missing index file. Please index the coord-sorted BAM file using `samtools index` command: {input_path}')
            bam_header = TriplixHeader(file_path=str(input_path))
            if bam_header.process_history.headers['HD']['SO'] != 'coordinate':
                raise ValueError(f'Input BAM file need to be "coordinate" sorted. Please sort using `samtools sort` command: {input_path}')

            # collecting interactions
            alignments = []
            n_ignore = 0
            with pysam.AlignmentFile(input_path, 'r', require_index=True) as bam_fid:
                alignment_iter = bam_fid.fetch(view_chrom, view_start, view_end)
                n_load = 0
                for aln_idx, aln in enumerate(alignment_iter):
                    if aln_idx % 1e6 == 0:
                        logger.debug(f'\t\tcurrently processing {aln_idx:d}th alignment')
                    if aln.mapping_quality < mapping_quality:
                        n_ignore += 1
                        continue
                    n_load += 1
                    alignment = dict(
                        id=aln.query_name,
                        chrom=aln.reference_name,
                        start=aln.reference_start,
                        end=aln.reference_end,
                        mapq=aln.mapping_quality,
                    )
                    alignments.append(alignment)
            logger.debug(f'\tIn total {n_load:,d} alignments are collected, and {n_ignore:,d} are ignored.')
            if n_load == 0:
                logger.warning(f'No alignments were found in this area, ignoring: {input_path}')
                continue

            # form a dataframe from alignments
            alignments = pd.DataFrame(alignments)
            alignments = alignments.sort_values(by=['id', 'start']).reset_index(drop=True)

            # excluding reads with #alignments < 2
            is_singleton = alignments.groupby(by='id', sort=False)['id'].transform('size') < 2
            alignments = alignments.loc[~is_singleton].reset_index(drop=True)
            logger.debug(f'\t{len(alignments):,d} alignments, and {len(alignments["id"].unique()):,d} reads are left after #alignment>1 filter')
            del is_singleton

            # map alignments to bins
            alignments['start_bidx'] = np.searchsorted(anchor_edges, alignments['start'], side='right') - 1
            alignments['end_bidx'] = np.searchsorted(anchor_edges, alignments['end'], side='right') - 1

            # add alignments to the counts
            n_dup = 0
            seen_ids = dict()
            n_contact = 0
            for read_idx, (read_id, read) in enumerate(alignments.groupby(by='id', sort=False)):
                n_frag = len(read)
                if n_frag < 1:
                    continue

                # store pairwise contacts
                for fi in range(n_frag):
                    for fj in range(fi + 1, n_frag):
                        frg_i = read.iloc[fi]
                        frg_j = read.iloc[fj]
                        # if np.abs(frg_j['start'] - frg_i['start']) < cli_args.min_dist:
                        #     continue

                        for ovl_i in range(frg_i['start_bidx'], frg_i['end_bidx'] + 1):
                            if not 0 <= ovl_i < n_anchor:
                                continue

                            for ovl_j in range(frg_j['start_bidx'], frg_j['end_bidx'] + 1):
                                if not 0 <= ovl_j < n_anchor:
                                    continue

                                pair_id = (ovl_i, ovl_j)
                                if pair_id not in seen_ids:
                                    seen_ids[pair_id] = set()
                                    if np.isnan(images['count_AB'][ovl_i, ovl_j]):
                                        images['count_AB'][ovl_i, ovl_j] = 0
                                        images['count_AB'][ovl_j, ovl_i] = 0

                                # ignore, if the read has already contributed to the current triplet
                                if read_id in seen_ids[pair_id]:
                                    n_dup += 1
                                    continue
                                seen_ids[pair_id].add(read_id)

                                images['count_AB'][ovl_i, ovl_j] += 1
                                images['count_AB'][ovl_j, ovl_i] += 1
                                n_contact += 1
            # seen_ids_collection[bam_fpath] = copy.deepcopy(seen_ids)
            logger.info(f'\t{n_contact:,d} contacts are stored, and {n_dup:,d} are ignored due to duplicity.')

        else:
            raise ValueError(f'Unknown file type: {input_path}')
    # n_load = np.nansum(image)
    # assert n_load > 0, 'No contacts were found in this area.'

    # prepare a figure
    # LinearSegmentedColormap.from_list('bwwr', ['#0000ff', '#ffffff', '#ff0000'])
    cmap = cm.get_cmap('hot_r', 20)
    cmap.set_bad(color='#b2b2a3', alpha=0.5)
    cmap.set_under(color='#ffffff', alpha=1.0)
    # cmap.set_over(color='#000055', alpha=1.0)
    ext_bnd = [view_start, view_end]
    n_row = int(np.ceil(len(plot_names) / 4))
    n_col = min(len(plot_names), 4)
    plt.close('all')
    fig_h, axes = plt.subplots(nrows=n_row, ncols=n_col, figsize=(n_col * 10, n_row * 8 + 1))
    axes = axes.flatten()
    for pi, col_name in enumerate(plot_names):
        axes[pi].img_h = axes[pi].imshow(
            images[col_name],
            interpolation='nearest',
            cmap=cmap, aspect='auto',
            extent=ext_bnd + [view_end, view_start],
        )

        axes[pi].cb_h = plt.colorbar(axes[pi].img_h, ax=axes[pi], fraction=0.025, pad=0.03, extend='both')
        # axes[pi].cb_h.ax.tick_params(labelsize=10)

        if col_name in plot_params and 'clim_range' in plot_params[col_name]:
            c_lim = [float(value) for value in plot_params[col_name]['clim_range'].split(',')]
        elif col_name in plot_params and 'clim_prct' in plot_params[col_name]:
            c_lim_prct = [float(value) for value in plot_params[col_name]['clim_prct'].split(',')]
            c_lim = [np.nanpercentile(images[col_name], c_lim_prct[0]), np.nanpercentile(images[col_name], c_lim_prct[1])]
        else:
            c_lim = [np.nanpercentile(images[col_name], 20), np.nanpercentile(images[col_name], 96)]
        axes[pi].img_h.set_clim(c_lim)

        x_tick_pos = np.unique(np.linspace(view_start, view_end, 25, dtype=int))
        axes[pi].set_xticks(x_tick_pos)
        axes[pi].set_yticks(x_tick_pos)
        axes[pi].xaxis.set_tick_params(rotation=60)
        axes[pi].xaxis.set_major_formatter(lambda v, _: '{:0.2f}'.format(v / 1e6))
        axes[pi].yaxis.set_major_formatter(lambda v, _: '{:0.2f}'.format(v / 1e6))

        axes[pi].set_title(f'{col_name}\nsum={np.nansum(images[col_name])}')

    plt.suptitle(f'{input_files[0].name}, #files used: {len(input_files)}\n'
                 f'{view_chrom}:{view_start / 1e6:0,.3f}-{view_end / 1e6:0,.3f}mb\n'
                 f'anchor width={anchor_width / 1e3:0.0f}k',
                 fontweight='normal')  # y=1 + 0.15 / (n_row * 3)
    plt.subplots_adjust(top=1.0 - 0.17 / n_row, wspace=0.2, hspace=0.2)  # , top=0.8
    # plt.show()

    logger_fonttools = logging.getLogger('fontTools.subset')
    loglevel_prev = logger_fonttools.level
    logger_fonttools.setLevel(logging.WARNING)
    plt.savefig(output_path, bbox_inches='tight')
    logger_fonttools.setLevel(loglevel_prev)
    plt.close()
    logger.info(f'Virtual-HiC plot is stored in: {output_path}')


def plot_processing_history(input_path, output_dir=None, output_name=None, **kwargs):
    input_path = pathlib.Path(input_path).expanduser()
    header = TriplixHeader(file_path=input_path)

    # prepare the output path
    if output_dir is None:
        output_dir = input_path.parent
    if output_name is None:
        output_name = input_path.stem + '.processing-history.pdf'
    output_path = pathlib.Path(output_dir).expanduser() / output_name
    output_path.parent.mkdir(parents=True, exist_ok=True)
    if output_path.is_file():
        output_path.unlink()

    header.process_history.visualize(output_path=output_path, **kwargs)


def plot_read_length_histogram(concatemers_path, output_dir=None, output_name=None):

    # prepare paths
    input_path = pathlib.Path(concatemers_path).expanduser()
    if output_dir is None:
        output_dir = input_path.parent
    if output_name is None:
        output_name = 'read-length.' + input_path.stem + '.pdf'
    output_path = pathlib.Path(output_dir).expanduser() / output_name
    output_path.parent.mkdir(parents=True, exist_ok=True)

    # prepare the container
    cctm_container = concatemers.ConcatemersContainer(input_path)
    concatemers_iter = cctm_container.iter(return_length=True)

    # define bin edges
    max_length = 10e3
    bin_edges = np.hstack([np.linspace(0, max_length, 101, dtype=int), np.inf])
    bin_width = bin_edges[1] - bin_edges[0]
    n_bin = len(bin_edges) - 1
    counts = np.zeros(n_bin, dtype=int)

    # iterate over concatemers
    logger.log_time = time.time()
    n_concatemer = cctm_container.h5_file['/reads/name'].shape[0]
    logger.info(f'Iterating over {n_concatemer:,d} concatemers in: {concatemers_path}')
    n_load = 0
    for cctm_idx, concatemer in enumerate(concatemers_iter):
        n_load += 1
        if time.time() - logger.log_time > configurations.configs['log_interval']:
            logger.log_time = time.time()
            logger.info(f'\t{n_load:,d}/{n_concatemer:,d} concatemers are processed. ')

        # collect read length
        bin_idx = np.searchsorted(bin_edges, concatemer['read_length'], side='right') - 1
        counts[bin_idx] += 1

    # prepare a figure
    plt.close('all')
    fig_h = plt.figure(figsize=(16, 5))
    ax = plt.gca()

    ax.bar(bin_edges[:-1], counts, width=bin_width * 0.9, color='#2f1ee6', align='edge')

    ax.xaxis.set_major_locator(ticker.MaxNLocator(nbins=23, integer=True))
    ax.yaxis.set_major_locator(ticker.MaxNLocator(nbins=10, integer=True))
    ax.xaxis.set_major_formatter(lambda v, _: '{:0,.0f}'.format(v))
    ax.yaxis.set_major_formatter(lambda v, _: '{:0,.0f}'.format(v))
    ax.set_xlim([0, max_length * 1.01])
    ax.set_xlabel('Read length (bases)')
    ax.set_ylabel('Number of reads')

    ax.set_title(
        f'Read length distribution:\n'
        f'{input_path.name}\n'
        f'#concatemers={n_load:,d}',
    )
    # plt.show()

    logger_fonttools = logging.getLogger('fontTools.subset')
    loglevel_prev = logger_fonttools.level
    logger_fonttools.setLevel(logging.WARNING)
    plt.savefig(output_path, bbox_inches='tight')
    logger_fonttools.setLevel(loglevel_prev)
    plt.close()
    logger.info(f'Read length histogram is stored in: {output_path}')


def plot_alignment_length_histogram(
        concatemers_path,
        output_dir=None, output_name=None,
        mapping_quality=1, max_length=3e3,
):
    max_length = int(max_length)

    # prepare paths
    input_path = pathlib.Path(concatemers_path).expanduser()
    if output_dir is None:
        output_dir = input_path.parent
    if output_name is None:
        output_name = 'alignment-length.' + input_path.stem + '.pdf'
    output_path = pathlib.Path(output_dir).expanduser() / output_name
    output_path.parent.mkdir(parents=True, exist_ok=True)

    # prepare the container
    cctm_container = concatemers.ConcatemersContainer(input_path)
    concatemers_iter = cctm_container.iter(return_length=True)

    # define bin edges
    bin_edges = np.hstack([np.linspace(0, max_length, 101, dtype=int), np.inf])
    bin_width = bin_edges[1] - bin_edges[0]
    n_bin = len(bin_edges) - 1
    counts = np.zeros(n_bin, dtype=int)

    # iterate over concatemers
    logger.log_time = time.time()
    n_concatemer = cctm_container.h5_file['/reads/name'].shape[0]
    logger.info(f'Iterating over {n_concatemer:,d} concatemers in: {concatemers_path}')
    n_load = 0
    n_alignments = 0
    for cctm_idx, concatemer in enumerate(concatemers_iter):
        n_load += 1
        if time.time() - logger.log_time > configurations.configs['log_interval']:
            logger.log_time = time.time()
            logger.info(f'\t{n_load:,d}/{n_concatemer:,d} concatemers are processed. ')

        # collect alignment lengths
        aln_lengths = concatemer['end'] - concatemer['start']
        bin_idxs = np.searchsorted(bin_edges, aln_lengths, side='right') - 1
        for aln_idx, bin_idx in enumerate(bin_idxs):
            if concatemer['map_quality'][aln_idx] < mapping_quality:
                continue
            n_alignments += 1
            counts[bin_idx] += 1

    # prepare a figure
    plt.close('all')
    fig_h = plt.figure(figsize=(16, 5))
    ax = plt.gca()

    ax.bar(bin_edges[:-1], counts, width=bin_width * 0.9, color='#2f1ee6', align='edge')

    ax.xaxis.set_major_locator(ticker.MaxNLocator(nbins=23, integer=True))
    ax.yaxis.set_major_locator(ticker.MaxNLocator(nbins=10, integer=True))
    ax.xaxis.set_major_formatter(lambda v, _: '{:0,.0f}'.format(v))
    ax.yaxis.set_major_formatter(lambda v, _: '{:0,.0f}'.format(v))
    ax.set_xlim([0, max_length * 1.01])
    ax.set_xlabel('Alignment length (bases)')
    ax.set_ylabel('Number of alignments')

    ax.set_title(
        f'Alignment length distribution:\n'
        f'{input_path.name}\n'
        f'#concatemers={n_load:,d}; #alignments={n_alignments:,d}; mapping_quality={mapping_quality:d}',
    )
    # plt.show()

    logger_fonttools = logging.getLogger('fontTools.subset')
    loglevel_prev = logger_fonttools.level
    logger_fonttools.setLevel(logging.WARNING)
    plt.savefig(output_path, bbox_inches='tight')
    logger_fonttools.setLevel(loglevel_prev)
    plt.close()
    logger.info(f'Alignment length histogram is stored in: {output_path}')





import sys
import gzip
import time
import pathlib
import logging
import hashlib
import importlib
import json
import re
# from dataclasses import dataclass
from typing import Union
import copy

import pysam
import h5py

from triplix.core import configurations
from triplix.core.utilities import get_random_string
from triplix.core.hdf5 import HDF5Container

logger = logging.getLogger(__name__)


class PGNode:
    def __init__(self, pg_info: Union[str, dict]):
        if isinstance(pg_info, str):
            assert pg_info.startswith('@PG')
            pg_info = dict(field.split(':', maxsplit=1) for field in pg_info.split('\t')[1:])
            if 'PS' in pg_info:
                if pg_info['PS'] == '':
                    pg_info['PS'] = []
                else:
                    pg_info['PS'] = pg_info['PS'].split(',')

        if 'VN' not in pg_info:
            pg_info['VN'] = 'NA'
        if 'CL' not in pg_info:
            pg_info['CL'] = ' '.join(sys.argv)
        if 'PS' not in pg_info:
            if 'PP' in pg_info:
                pg_info['PS'] = [pg_info['PP']]
            else:
                pg_info['PS'] = []
        assert isinstance(pg_info['PS'], list), 'PS field should be a list of parent(s)'
        # else:
        #     if isinstance(pg_info['PS'], str):
        #         pg_info['PS'] = [pg_info['PS']]

        self.data = pg_info
        self.children_ids = []

    def hash(self):
        hasher = hashlib.md5()
        hasher.update(json.dumps(self.data, sort_keys=True).encode())
        return hasher.hexdigest()

    def __getitem__(self, field_name: str):
        return self.data[field_name]

    def __setitem__(self, field_name: str, value: str):
        self.data[field_name] = value

    def __contains__(self, item):
        return item in self.data

    def __str__(self):
        out = ['@PG:']
        for key, value in self.data.items():
            if isinstance(value, str):
                out += [f'{key}:{value}']
            else:
                out += [f'{key}:{",".join(value)}']
        return '\t'.join(out)

    def __repr__(self):
        out = []
        for key, value in self.data.items():
            out += [f'{key}> {value}']
        for child_id in self.children_ids:
            out += [f'Child --> {child_id}']
        return '\n'.join(out)


class ProcessHistoryDAG:

    def __init__(self, sam_headers=None):
        self.headers = {'HD': {}, 'SQ': []}
        self.programs = {}
        self.source_ids = []
        self.destination_id = None
        if sam_headers is not None:
            self.from_sam_headers(sam_headers)

    def from_sam_headers(self, sam_headers):
        for hdr_idx, sam_header in enumerate(sam_headers):
            if sam_header[:3] in ['@HD', '@SQ']:
                fields = sam_header.split('\t')
                values = dict(field.split(':', maxsplit=1) for field in fields[1:])
                if fields[0] == '@HD':
                    self.headers['HD'] = values
                else:
                    self.headers['SQ'].append(values)
            else:
                assert sam_header.startswith('@PG')
                pg_node = PGNode(sam_header)
                self.add_pg(pg_node)

    def to_sam_headers(self):

        # sam headers
        sam_headers = [
            f'@HD\t' +
            '\t'.join([f'{key}:{value}' for key, value in self.headers['HD'].items()])
        ]
        for sq_item in self.headers['SQ']:
            sam_headers.append(
                f'@SQ\t' +
                '\t'.join([f'{key}:{value}' for key, value in sq_item.items()])
            )

        # processing history
        for program_id, program in self.programs.items():
            row = ['@PG:']
            for key, value in program.data.items():
                if isinstance(value, str):
                    row += [f'{key}:{value}']
                elif key in ['PS']:
                    if len(value) > 0:
                        row += [f'{key}:{",".join(value)}']
                else:
                    row += [f'{key}:{",".join(value)}']
            sam_headers += ['\t'.join(row)]
        return sam_headers

    def add_pg(self, pg_node: Union[PGNode, dict], is_destination=True):
        if isinstance(pg_node, dict):
            pg_node = PGNode(pg_node)
        assert pg_node['ID'] not in self.programs
        self.programs[pg_node['ID']] = pg_node

        if len(pg_node['PS']) == 0:
            self.source_ids.append(pg_node['ID'])
        else:
            for parent_id in pg_node['PS']:
                assert parent_id in self.programs
                self.programs[parent_id].children_ids.append(pg_node['ID'])

        if is_destination:
            self.destination_id = pg_node['ID']

    def __repr__(self):
        if len(self.source_ids) == 0:
            return 'Empty History'
        out = ''
        for source_id in self.source_ids:
            out += f'--- Source: {self.programs[source_id]["ID"]}\n'
            out += self._print_children(self.programs[source_id], level=0)
            out += f'--- End of: {self.programs[source_id]["ID"]}\n'
        return out

    def _print_children(self, pg_node: PGNode, level=0):
        out = "|\t" * level + f'{pg_node}\n'
        for child_id in pg_node.children_ids:
            out += self._print_children(self.programs[child_id], level=level + 1)
        return out

    def visualize(self, output_path=None, **kwargs):
        igraph_installed = importlib.util.find_spec('igraph') is not None
        assert igraph_installed, '"igraph" library is not installed. Please install it using: `pip install igraph`'
        import igraph
        from matplotlib import pyplot as plt
        import matplotlib as mpl

        # generate the graph
        ph_graph = igraph.Graph(directed=True)
        ph_graph.add_vertices(list(self.programs.keys()))
        ph_graph.vs['label'] = ph_graph.vs['name']
        ph_graph.vs['color'] = '#4fd1fc'
        root_idxs = []
        for pg_idx, (pg_name, pg_node) in enumerate(self.programs.items()):

            # annotate the source nodes
            if len(pg_node['PS']) == 0:
                ph_graph.vs[pg_idx]['color'] = '#81fd90'
                root_idxs.append(pg_idx)

            # adding edges
            for child_id in pg_node.children_ids:
                ph_graph.add_edge(pg_name, child_id)
            # if ph_graph.ecount() > 20:
            #     break

        # filtering isolated vertices
        ph_graph.vs.select(_degree=0).delete()

        # plt.close('all')
        fig_h = plt.figure(figsize=kwargs.get('figsize', (10, 7)))
        ax = plt.gca()

        # graph_layout = ph_graph.layout('kamada_kawai')
        # graph_layout = ph_graph.layout(kwargs.get('layout', 'reingold_tilford'))
        # graph_layout = ph_graph.layout_reingold_tilford(root=root_idxs)
        # graph_layout = ph_graph.layout_reingold_tilford_circular(root=root_idxs)
        graph_layout = ph_graph.layout_sugiyama()
        # graph_layout = ph_graph.layout_davidson_harel()
        ax.img_h = igraph.plot(
            ph_graph,
            target=ax,
            layout=graph_layout,
            vertex_shape='rectangle',
            vertex_label=ph_graph.vs['name'],
            vertex_label_size=kwargs.get('fontsize', 4),
            vertex_size=15,
            vertex_color=ph_graph.vs['color'],
            vertex_frame_color='none',
            alpha=1.0,
            **kwargs,
        )
        # possible shapes:
        #         "rectangle": "s", # always a square
        #         "circle": "o",
        #         "hidden": "none",
        #         "triangle-up": "^",
        #         "triangle-down": "v",
        # shape_size = [len(label) * 2 for label in ph_graph.vs['label']]
        # text_size = np.array([len(label) for label in ph_graph.vs['name']])
        # label_dist = (text_size - text_size.min()) / (text_size.max() - text_size.min())
        ax.invert_yaxis()
        plt.axis('off')
        plot_title = kwargs.get('title', None)
        if plot_title is not None:
            ax.set_title(plot_title, fontweight='bold')
            # plt.suptitle(plot_title, fontweight='bold')
        # fig_h.set_visible(not fig_h.get_visible())
        # plt.draw()

        # align text objects
        for child in ax.get_children():
            if isinstance(child, mpl.text.Text):
                child.set_horizontalalignment('center')
                child.set_verticalalignment('center')
        # todo: adding a rectangle for each node, to make its size adjustable
        # ax.collections[0].get_offsets().data

        # store the graph if requested
        if output_path is None:
            plt.show()
        else:
            output_path = pathlib.Path(output_path).expanduser()

            logger_fonttools = logging.getLogger('fontTools.subset')
            loglevel_prev = logger_fonttools.level
            logger_fonttools.setLevel(logging.WARNING)
            plt.savefig(fname=output_path)  # , bbox_inches='tight'
            logger_fonttools.setLevel(loglevel_prev)
            plt.close()
            logger.info(f'The processing history plot is stored in: {output_path}')


class TriplixHeader:
    FORMAT_VERSION = '1.0.0'

    def __init__(self, file_path=None, exclusive=False):
        self.exclusive = exclusive

        self.format_version = TriplixHeader.FORMAT_VERSION
        self.triplix_version = configurations.configs['version']
        self.creation_timestamp = time.strftime("%Y-%m-%d %H:%M:%S")
        self.cell_type = None
        self.assay_name = None
        self.experiment_name = None
        self.source_file = None
        self.assembly_name = None
        self.assembly_length = None
        self.chrom_lengths = {}
        self.stored_chunks = []
        self.shape = 'upper triangle'
        self.sort_order = []
        self.anchor_width = None
        self.transformations = []
        self.background_normalizations = {}
        # self.sam_headers = []
        self.column_names = []
        self.process_history = ProcessHistoryDAG()

        # todo: add column_dtypes to headers
        # self.column_dtypes = []

        self.file_path = file_path
        if self.file_path is not None:
            self.extract_header(self.file_path)

    def __repr__(self):
        return self.to_string()

    def extract_header(self, file_path):
        self.file_path = pathlib.Path(file_path).expanduser()
        assert self.file_path.is_file(), f'File not found: {self.file_path}'
        self.file_path = str(self.file_path)

        # extract using appropriate method
        if self.file_path.endswith('.bam'):
            _verbosity_level = pysam.set_verbosity(0)
            with pysam.AlignmentFile(self.file_path, 'r') as bam_file:
                pysam.set_verbosity(_verbosity_level)
                sam_headers = str(bam_file.header).rstrip().split('\n')
                self.process_history = ProcessHistoryDAG(sam_headers)
                self.chrom_lengths = {}
                for sq_item in self.process_history.headers['SQ']:
                    self.chrom_lengths[sq_item['SN']] = int(sq_item['LN'])
        elif self.file_path.endswith('.concatemers.h5'):
            logger.debug(f'Opening (mode=r): {self.file_path}')
            with h5py.File(self.file_path, 'r') as h5_file:
                cnct_grp = h5_file['concatemers']
                self.format_version = cnct_grp.attrs['format_version']
                self.triplix_version = cnct_grp.attrs['triplix_version']
                self.creation_timestamp = cnct_grp.attrs['creation_timestamp']
                self.assay_name = cnct_grp.attrs['assay_name']
                self.cell_type = cnct_grp.attrs['cell_type']
                self.experiment_name = cnct_grp.attrs['experiment_name']
                self.source_file = cnct_grp.attrs['source_file']
                if 'genome_name' in cnct_grp.attrs:
                    self.assembly_name = cnct_grp.attrs['genome_name']
                else:
                    self.assembly_name = cnct_grp.attrs['assembly_name']
                if 'genome_length' in cnct_grp.attrs:
                    self.assembly_length = cnct_grp.attrs['genome_length']
                else:
                    self.assembly_length = cnct_grp.attrs['assembly_length']
                self.chrom_lengths = dict(zip(
                    cnct_grp.attrs['chrom_names'],
                    cnct_grp.attrs['chrom_lengths'],
                ))
                stored_chunks = [re.split(r'[:-]', range_str) for range_str in cnct_grp.attrs['stored_chunks']]
                self.stored_chunks = [[chrom, int(start), int(end)] for (chrom, start, end) in stored_chunks]
                self.shape = 'upper triangle'
                self.sort_order = []
                sam_headers = cnct_grp.attrs['sam_headers'].tolist()
                self.process_history = ProcessHistoryDAG(sam_headers)
                self.column_names = [
                    'chrom_A', 'start_A', 'end_A',
                    'chrom_B', 'start_B', 'end_B',
                    'chrom_C', 'start_C', 'end_C',
                    'strand_A', 'strand_B', 'strand_C',
                    'mapq_A', 'mapq_B', 'mapq_C',
                    'frag_index_A', 'frag_index_B', 'frag_index_C',
                    'read_name', 'read_length_nbp', 'read_length_nfrag',
                    'experiment_name',
                ]
        elif self.file_path.endswith(('.tri-alignments.tsv.bgz', '.btrp')):
            with gzip.open(self.file_path, 'rt') as gz_file:

                # get file type
                file_type = next(gz_file).rstrip()
                assert file_type in ['## Triplets', '## Triplix'], 'Unknown file format'

                sam_headers = []
                for line in gz_file:
                    line = line.rstrip()
                    if not line.startswith('#'):
                        raise ValueError('Unknown header format')

                    # process key, values
                    key, attr_value = line[1:].split(': ')
                    attr_name = key.replace(' ', '_')
                    if attr_name in ['sort_order', 'column_names']:
                        setattr(self, attr_name, attr_value.split('\t'))
                        if attr_name == 'sort_order' and len(attr_value) == 0:
                            setattr(self, 'sort_order', [])
                    elif attr_name == 'transformation':
                        transformation = dict(field.split(':') for field in attr_value.split('\t'))
                        self.transformations.append(transformation)
                    elif attr_name == 'chrom_lengths':
                        self.chrom_lengths = {}
                        for chr2len_pair in attr_value.split('\t'):
                            chr_name, chr_len = chr2len_pair.split(':')
                            self.chrom_lengths[chr_name] = int(chr_len)
                    elif attr_name in ['stored_chunks', 'table_chunks']:  # todo: 'table_chunks' should be removed, added for compatibility
                        self.stored_chunks = []
                        for chr2rng_pair in attr_value.split('\t'):
                            chr_name, chr_rng = chr2rng_pair.split(':')
                            chr_start, chr_end = [int(item) for item in chr_rng.split('-')]
                            self.stored_chunks.append([chr_name, chr_start, chr_end])
                    elif attr_name == 'sam_header':
                        sam_headers.append(attr_value)
                    elif attr_name in ['resolution', 'anchor_width', 'genome_length', 'assembly_length']:
                        # if attr_name == 'anchor_width':
                        #     logger.warning('Anchor width is ignored')
                        #     setattr(self, attr_name, None)
                        #     continue
                        setattr(self, attr_name, int(attr_value))
                    else:
                        setattr(self, attr_name, attr_value)
                    if attr_name == 'column_names':
                        break
            self.process_history = ProcessHistoryDAG(sam_headers)
        elif self.file_path.endswith('.triplets.h5'):
            with HDF5Container(self.file_path, mode='r', exclusive=self.exclusive) as container:
                expr_grp = container.h5_file['experiments']
                self.experiment_name = list(expr_grp.keys())[0]
                self.anchor_width = expr_grp[self.experiment_name].attrs['anchor_width']

                chrom_grp = expr_grp[self.experiment_name]['chroms']
                self.chrom_lengths = {chr_name.decode(): chr_length for chr_name, chr_length in zip(chrom_grp['name'], chrom_grp['length'])}
        else:
            raise ValueError('Unknown file extension.')

    def to_string(self):
        outputs = ['## Triplix']
        headers_ordered = [
            'format version',
            'triplix version',
            'creation timestamp',
            'cell type',
            'assay name',
            'experiment name',
            'source file',
            'assembly name',
            'assembly length',
            'chrom lengths',
            'stored chunks',
            'shape',
            'sort order',
            'anchor width',
            'transformations',
            'background normalizations',
            'sam headers',
            'column names',
        ]

        # iterate over headers and add their string representation
        for header in headers_ordered:
            attr_name = header.replace(' ', '_')
            if attr_name in ['sam_headers']:
                attr_value = self.process_history.to_sam_headers()
            else:
                if not hasattr(self, attr_name):
                    logger.warning(f'Missing attribute: {attr_name}')
                    continue
                attr_value = getattr(self, attr_name)
            if header == 'chrom lengths':
                chrom_lengths_str = [f'{chr_name}:{chr_len:d}' for chr_name, chr_len in self.chrom_lengths.items()]
                outputs += [f'#chrom lengths: ' + '\t'.join(chrom_lengths_str)]
            elif header == 'sam headers':
                for sam_header in self.process_history.to_sam_headers():
                    outputs += ['#sam header: ' + sam_header]
            elif header == 'stored chunks':
                stored_chunks = [f'{chr_name}:{start:d}-{end:d}' for chr_name, start, end in self.stored_chunks]
                outputs += [f'#stored chunks: ' + '\t'.join(stored_chunks)]
            elif header == 'transformations':
                for transformation in self.transformations:
                    outputs += [f'#transformation: ' + '\t'.join(f'{key}:{value}' for key, value in transformation.items())]
            elif header == 'background normalizations':
                if len(self.background_normalizations) > 0:
                    outputs += [f'#background normalizations: ' + '\t'.join(f'{feat_name}:{norm["name"]}' for feat_name, norm in self.background_normalizations.items())]
            elif isinstance(attr_value, list):
                if len(attr_value) != 0:  # second-level if is needed
                    outputs += [f'#{header}: ' + '\t'.join(attr_value)]
            elif attr_value is not None:
                outputs += [f'#{header}: {attr_value}']

        return '\n'.join(outputs)

    def update(self, **kwargs):
        for attr, value in kwargs.items():
            setattr(self, attr, value)

    def merge_process_histories(self, headers, check_uniqueness=False):
        assert len(headers) > 0
        self.process_history = ProcessHistoryDAG()
        self.process_history.headers['HD'] = headers[0].process_history.headers['HD']
        self.process_history.headers['SQ'] = headers[0].process_history.headers['SQ']

        # validate HD rows, they all should be identical
        for header in headers:
            if self.process_history.headers['HD'] != header.process_history.headers['HD']:
                raise SyntaxError(
                    f'Inconsistent @HD lines:\n'
                    f'      One @HD is: {self.process_history.headers["HD"]}\n'
                    f'The other @HD is: {header.process_history.headers["HD"]}'
                )

        # validate SQ rows, they all should be identical. The order is kept
        for seq_info in self.process_history.headers['SQ']:
            for header in headers:
                try:
                    header.process_history.headers['SQ'].index(seq_info)
                except ValueError:
                    raise ValueError(f'Inconsistent @SQ: Missing {seq_info}')

        # merge process histories
        for header in headers:
            for pg_id, pg_node in header.process_history.programs.items():
                if pg_node['ID'] in self.process_history.programs:
                    if check_uniqueness:
                        raise ValueError(
                            f'Duplicated PG:ID is found:\n'
                            f'Prog 1: {self.process_history.programs[pg_node["ID"]]}\n'
                            f'Prog 2: {pg_node}'
                        )
                else:
                    self.process_history.programs[pg_node['ID']] = pg_node
                    if len(pg_node['PS']) == 0:
                        self.process_history.source_ids.append(pg_node['ID'])
        self.process_history.destination_id = None




import pathlib
import logging
import tempfile
from typing import Union

import h5py
import fasteners
import numpy as np

from triplix.core import configurations

COMMAND_NAME = 'triplix.hdf5'
logger = logging.getLogger(COMMAND_NAME)


class DatasetIterator:

    def __init__(self, hdf5_path: Union[str, h5py.File], dataset_path):
        if isinstance(hdf5_path, h5py.File):
            self.h5_handle = hdf5_path
            logger.debug(f'Using the provided file handle: {self.h5_handle}')
        else:
            self.hdf5_path = pathlib.Path(hdf5_path).expanduser()
            assert self.hdf5_path.is_file(), f'File not found: {self.hdf5_path}'
            logger.debug(f'Opening (mode=r): {self.hdf5_path}')
            self.h5_handle = h5py.File(self.hdf5_path, mode='r')

        logger.debug(f'Selected dataset is: {dataset_path}')
        self.dataset = self.h5_handle[dataset_path]
        self.dataset_path = dataset_path
        self.dataset_iter_chunks = self.dataset.iter_chunks()
        self.data = self.dataset[next(self.dataset_iter_chunks)]
        self.offset = 0
        self.cursor = 0

    def get_next(self, n_items=1):

        # do we have enough loaded data?
        while self.cursor + n_items > len(self.data):

            # delete old (used) data
            if self.cursor != 0:
                self.data = self.data[self.cursor:]
                self.offset += self.cursor
                self.cursor = 0

            # collect new chunk of data
            try:
                chunk_slice = next(self.dataset_iter_chunks)
            except StopIteration:
                raise StopIteration(f'End of the dataset is reached: '
                                    f'dataset shape: {self.dataset.shape}, '
                                    f'requested index: {self.offset + n_items}.')
            self.data = np.concatenate((self.data, self.dataset[chunk_slice]), axis=0)

        start_idx = self.cursor
        self.cursor += n_items
        return self.data[start_idx: self.cursor]

    def __iter__(self):
        return self

    def __next__(self):
        return self.get_next(n_items=1)[0]

    def __len__(self):
        return self.dataset.shape[0]

    def __repr__(self):
        return (
            f'DatasetIterator: '
            f'path={self.dataset_path}, '
            f'shape={self.dataset.shape} '
            f'cursor={self.offset + self.cursor}'
        )


class HDF5Container:

    def __init__(self, container_path, mode='a', exclusive=False):

        # initializations
        self.container_path = pathlib.Path(container_path).expanduser()
        self.lock_handle = fasteners.InterProcessLock(f'{self.container_path}.lock')  # using temp folder wont work: in a cluster, the temp folder could be different for each node
        self.exclusive = exclusive
        self.dataset_defaults = dict(
            maxshape=(None,),
            chunks=(configurations.configs['concatemers']['hdf5_chunk_size'],),
            compression="gzip",
            compression_opts=configurations.configs['concatemers']['compression_level'],
            # track_order=True, # could cause problems if true, an HDF5 bug
        )

        # opening the file handle
        if self.exclusive:
            self.lock()
        logger.debug(f'Opening (mode={mode}): {self.container_path}')
        self.h5_file = h5py.File(self.container_path, mode=mode)

    def lock(self):
        logger.debug(f'Requesting (an exclusive) right to edit: {self.container_path}')
        self.lock_handle.acquire()
        logger.debug(f'Permission acquired: {self.container_path}')

    def unlock(self):
        self.lock_handle.release()
        lock_path = self.lock_handle.path.decode()
        # pathlib.Path(lock_path).unlink()
        logger.debug(f'Container lock is released: {lock_path}')

    def store(self, dataset_path, row_indices, values, exclusive=True):
        if exclusive:
            self.lock()
        assert dataset_path in self.h5_file, f'Dataset not found: {dataset_path}'
        assert not hasattr(values, '__iter__') or len(row_indices) == len(values)

        # reshape if needed
        max_index = np.max(row_indices)
        if max_index >= self.h5_file[dataset_path].shape[0]:
            logger.warning(f'Dataset size is smaller than input index. Increasing size of {dataset_path} '
                           f'from {self.h5_file[dataset_path].shape[0]} to {max_index}')
            self.h5_file[dataset_path].resize(max_index + 1, axis=0)

        # store
        self.h5_file[dataset_path][row_indices] = values
        if exclusive:
            self.unlock()

    def retrieve(self, dataset_path, row_indices, exclusive=True):
        if exclusive:
            self.lock()
        assert dataset_path in self.h5_file, f'Dataset not found: {dataset_path}'

        # check if out-of-bound indices exist
        max_index = np.max(row_indices)
        n_row = self.h5_file[dataset_path].shape[0]
        if max_index >= n_row:
            logger.warning(f'{dataset_path} (#row={n_row:,d}): Out-of-bound indices (maximum_index={max_index:,d}). These indices will be ignored.')
            row_indices = row_indices[row_indices < n_row]

        # retrieve
        values = self.h5_file[dataset_path][row_indices]
        if exclusive:
            self.unlock()
        return values

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        logger.debug(f'Closing (mode={self.h5_file.mode}): {self.h5_file.filename}')
        self.h5_file.close()
        if self.exclusive:
            self.unlock()



import copy
import pathlib
import re
import time
from typing import Any, Union, Iterable
import logging
from collections import deque

import fasteners
import h5py
import numpy as np
import pandas as pd
from triplix.core import configurations
from triplix.core.hdf5 import HDF5Container
from triplix.core.header import TriplixHeader
from triplix.core.utilities import overlap, generate_prog_id, merge_intervals

__version__ = '1.0.0'
COMMAND_NAME = 'triplix.triplets'
COMMAND_ID = f'{COMMAND_NAME}>{generate_prog_id()}'
logger = logging.getLogger(COMMAND_NAME)


class Triplets(dict):

    def __init__(self, *args, **kwargs):
        columns = kwargs.pop('columns', [])

        # initialize the `dict` object
        super().__init__(*args, **kwargs)
        self.options = {
            'display.min_rows': 5
        }

        # initialize columns, if given
        for column in columns:
            self[column] = []

    def __repr__(self):
        # if 'start_A' not in self:
            # return '<Miss-formatted Triplets>: Missing `start_A` column'
        n_row = len(self)
        if n_row == 0:
            return f'--- Empty Triplets ---'
        
        n_show = self.options['display.min_rows']
        if n_row <= n_show:
            show_df = pd.DataFrame(self).astype(str)
        else:
            show_idxs = list(range(0, n_show + 1)) + list(range(n_row - n_show, n_row))
            show_df = pd.DataFrame({key: np.array(values)[show_idxs] for (key, values) in self.items()}, index=show_idxs)
            show_df = show_df.astype(str)
            show_df.loc[n_show, :] = '***'
            show_df.index = show_df.index.astype(str)
            show_df.rename({str(n_show): '***'}, axis=0, inplace=True)
        return f'Triplets <{n_row} x {len(self.keys())}>:\n' + repr(show_df)

    def __len__(self):
        for column in self.keys():
            return len(self[column])
        return 0

    def __getitem__(self, pointer: Union[str, slice, int, Iterable]):
        if isinstance(pointer, str):  # we are requesting a key/column of the parent dict object
            return dict.__getitem__(self, pointer)

        out = Triplets()
        for col in self.keys():
            array = np.array(dict.__getitem__(self, col))

            # selecting the requested rows
            if isinstance(pointer, (slice, )):
                out[col] = array[pointer]
            elif hasattr(pointer, '__iter__'):  # or more specifically: isinstance(pointer, (list, pd.Series, np.ndarray))
                out[col] = array[pointer]
            elif isinstance(pointer, int):
                out[col] = [array[pointer]]
            else:
                raise ValueError(f'Unknown index type: {pointer}')
        return out

    def clear(self):
        for column in self.keys():
            self[column] = []

    def to_dataframe(self, *args, **kwargs):
        return pd.DataFrame(self, *args, **kwargs)

    def to_slice(self, view_point, anchor_edges, value_column='count_ABC', symmetric=True):

        # find viewpoint's index
        vp_idx = np.searchsorted(anchor_edges, view_point, side='right') - 1
        anchor_idxs = np.searchsorted(anchor_edges, [
            self['start_A'],
            self['start_B'],
            self['start_C']
        ], side='right').T - 1

        # sort indices relative to viewpoint
        vp_dist = np.abs(anchor_idxs - vp_idx)
        sorted_idxs = np.take_along_axis(anchor_idxs, np.argsort(vp_dist, axis=1), axis=1)

        # find triplets having a viewpoint
        n_anchor = len(anchor_edges) - 1
        has_vp = sorted_idxs[:, 0] == vp_idx
        is_valid = (
            has_vp &
            (anchor_idxs >= 0).all(axis=1) &
            (anchor_idxs < n_anchor).all(axis=1)
        )

        # populate the map
        slice_map = np.full([n_anchor, n_anchor], fill_value=np.nan)
        slice_map[sorted_idxs[is_valid, 1], sorted_idxs[is_valid, 2]] = self[value_column][is_valid]
        if symmetric:
            slice_map[sorted_idxs[is_valid, 2], sorted_idxs[is_valid, 1]] = self[value_column][is_valid]

        return slice_map

    def to_cube(self, anchor_edges, value_column='count_ABC'):

        # find triplets overlapping the given anchor_edges
        anchor_idxs = np.searchsorted(anchor_edges, [self['start_A'], self['start_B'], self['start_C']], side='right').T - 1
        n_anchor = len(anchor_edges) - 1
        is_valid = (
                np.all(anchor_idxs >= 0, axis=1) &
                np.all(anchor_idxs < n_anchor, axis=1)
        )

        # populate the cube of counts
        cube_map = np.full([n_anchor, n_anchor, n_anchor], fill_value=np.nan)
        cube_map[
            anchor_idxs[is_valid, 0],
            anchor_idxs[is_valid, 1],
            anchor_idxs[is_valid, 2]
        ] = self[value_column]

        return cube_map

    def sort_anchors(self, by, force_upper=True):
        by = int(by)
        triplets = copy.deepcopy(self)

        sorted_idxs = None
        for anchor_type in ['start', 'end']:
            if f'{anchor_type}_A' not in triplets:
                continue
            coords = np.array([
                triplets[f'{anchor_type}_A'],
                triplets[f'{anchor_type}_B'],
                triplets[f'{anchor_type}_C'],
            ], dtype=int).T
            if anchor_type == 'start':
                anchor_distance = np.abs(coords - by)
                sorted_idxs = np.argsort(anchor_distance, axis=1)

            sorted_coords = np.take_along_axis(coords, sorted_idxs, axis=1)
            triplets[f'{anchor_type}_A'] = sorted_coords[:, 0]
            triplets[f'{anchor_type}_B'] = sorted_coords[:, 1]
            triplets[f'{anchor_type}_C'] = sorted_coords[:, 2]

        if force_upper:
            is_tril = triplets['start_B'] > triplets['start_C']
            if np.any(is_tril):
                triplets['start_B'][is_tril], triplets['start_C'][is_tril] = triplets['start_C'][is_tril], triplets['start_B'][is_tril]
                if 'end_B' in triplets:
                    triplets['end_B'][is_tril], triplets['end_C'][is_tril] = triplets['end_C'][is_tril], triplets['end_B'][is_tril]

        return triplets


class TripletsContainer:

    def __init__(self, container_path, tri_alignments_path=None, anchor_width=None, anchor_max_distance=None, exclusive=False, auto_init=False):
        self.container_path = pathlib.Path(container_path).expanduser()
        self.dataset_defaults = dict(
            maxshape=(None,),
            chunks=(configurations.configs['triplets']['hdf5_chunk_size'],),
            compression="gzip",
            compression_opts=configurations.configs['triplets']['compression_level'],
        )

        # prepare the lock
        self.lock_handle = fasteners.InterProcessLock(f'{self.container_path}.lock')
        self.exclusive = exclusive

        # check if the container needs initialization
        self.properties = {}
        if auto_init:
            logger.debug(f'Checking if the container is present: {self.container_path}')
            self.lock()
            if self.container_path.is_file():
                logger.debug('The input Triplets container is present. There is no need for initialization.')
            else:
                logger.debug('No Triplet container is found. Initializing the container.')
                assert tri_alignments_path is not None, 'When initializing the container, `tri_alignments_path` argument is required.'
                assert anchor_width is not None, 'When initializing the container, `anchor_width` argument is required.'
                assert anchor_max_distance is not None, 'When initializing the container, `anchor_max_distance` argument is required.'
                self.properties = {
                    'anchor_max_distance': int(anchor_max_distance),
                    'anchor_width': int(anchor_width),
                }
                self.initialize_container(tri_alignments_path=tri_alignments_path, exclusive=False)
            self.unlock()

        # collect experiment properties
        logger.debug('Collecting experiment properties')
        self.fetch_properties(exclusive=True)  # should be exclusive, otherwise it fails when multiple process is reading from it
        # one solution could be: https://stackoverflow.com/questions/68457924/error-while-accessing-hdf5-file-shows-error-oserror-unable-to-open-file
        self.experiment_path = f'experiments/{self.properties["experiment_name"]}/'

        if anchor_width is not None:
            assert self.properties['anchor_width'] == anchor_width, f'Inconsistent argument: `anchor_width`={anchor_width}'
        if anchor_max_distance is not None:
            assert self.properties['anchor_max_distance'] == anchor_max_distance, f'Inconsistent argument: `anchor_max_distance`={anchor_max_distance}'
        if self.properties['storage_format'] != 'full':
            raise NotImplementedError('Only "full" storage format is supported')
        if self.properties['anchor_type'] != 'fixed':
            raise NotImplementedError('Only fixed-width anchors is supported')

        if self.exclusive:
            self.lock()

    def __enter__(self):
        return self
    
    def __exit__(self, exception_type, exception_value, exception_traceback):
        if self.exclusive:
            self.unlock()
    
    def iter_chunks(self, chroms=None, columns=None):
        if chroms is None:
            chroms = self.properties['chrom_names']
        if columns is None:
            columns = self.properties['columns_order']
        if not isinstance(chroms, list):
            chroms = [chroms]
        if not isinstance(columns, list):
            columns = [columns]

        with HDF5Container(self.container_path, mode='r', exclusive=False) as container:
            for chrom in chroms:
                cis_path = f'{self.experiment_path}/contacts/triplets/{chrom},{chrom},{chrom}'
                if cis_path not in container.h5_file:
                    continue
                
                # preparing a chunk iterator
                cis_group = container.h5_file[cis_path]
                ds_slicer = cis_group[columns[0]].iter_chunks()
                
                # iterating over chunks
                triplets = Triplets()
                while True:
                    try:
                        ds_slice = next(ds_slicer)
                        for column in columns:
                            triplets[column] = cis_group[column][ds_slice]
                        yield chrom, triplets
                    except StopIteration:
                        break
    
    def to_dataframe(self, *args, **kwargs):
        return pd.DataFrame(self, *args, **kwargs)
    
    def lock(self):
        logger.debug(f'Requesting (an exclusive) right to: {self.container_path}')
        self.lock_handle.acquire()
        logger.debug(f'Permission acquired: {self.container_path}')

    def unlock(self):
        self.lock_handle.release()
        lock_path = self.lock_handle.path.decode()
        # pathlib.Path(lock_path).unlink()
        logger.debug(f'Container lock is released: {lock_path}')

    def initialize_container(self, tri_alignments_path, exclusive=True):
        trialn_header = TriplixHeader(file_path=tri_alignments_path)
        anchor_width = self.properties['anchor_width']
        anchor_max_distance = self.properties['anchor_max_distance']

        with HDF5Container(self.container_path, mode='w', exclusive=exclusive) as container:

            # experiment attributes
            expr_grp = container.h5_file.create_group(f'/experiments/{trialn_header.experiment_name}')  # , track_order=True: keep False, because of https://github.com/h5py/h5py/issues/1385
            expr_grp.attrs['format'] = 'HDF5::Triplix_Triplets'
            expr_grp.attrs['format_version'] = __version__
            expr_grp.attrs['triplix_version'] = configurations.configs['version']
            expr_grp.attrs['creation_timestamp'] = time.strftime("%Y-%m-%d %H:%M:%S")
            expr_grp.attrs['assembly_name'] = trialn_header.assembly_name
            expr_grp.attrs['assembly_length'] = sum(trialn_header.chrom_lengths.values())
            if trialn_header.assay_name is not None:
                expr_grp.attrs['assay_name'] = trialn_header.assay_name
            else:
                expr_grp.attrs['assay_name'] = 'unknown'
            if trialn_header.cell_type is not None:
                expr_grp.attrs['cell_type'] = trialn_header.cell_type
            else:
                expr_grp.attrs['cell_type'] = 'unknown'
            expr_grp.attrs['experiment_name'] = trialn_header.experiment_name
            expr_grp.attrs['source_bam_file'] = str(trialn_header.source_file)
            expr_grp.attrs['storage_format'] = 'full'  # {'full', 'sparse'}
            expr_grp.attrs['storage_mode'] = 'symmetric, upper triangle'  # { "symmetric, upper triangle", "square" }
            expr_grp.attrs['anchor_type'] = 'fixed'  # { "fixed", "variable" }
            expr_grp.attrs['anchor_width'] = int(anchor_width)
            expr_grp.attrs['anchor_max_distance'] = int(anchor_max_distance)
            expr_grp.attrs['user_metadata'] = '{}'
            if expr_grp.attrs['storage_format'] != 'full':
                raise NotImplementedError('Only "full" storage format is supported')
            if expr_grp.attrs['anchor_type'] != 'fixed':
                raise NotImplementedError('Only fixed-width anchor is supported')

            # add meta-data group
            trialn_header.update(
                creation_timestamp=time.strftime("%Y-%m-%d %H:%M:%S"),
                format_version=TriplixHeader.FORMAT_VERSION,
                anchor_width=anchor_width,
                column_names=[],
            )
            meta_grp = expr_grp.create_group(f'metadata/')
            meta_grp.create_dataset(name='triplix_headers', data=trialn_header.to_string().split('\n'), dtype=h5py.string_dtype(encoding='utf-8'), **self.dataset_defaults)
            meta_grp.create_dataset(name='sam_headers', data=trialn_header.process_history.to_sam_headers(), dtype=h5py.string_dtype(encoding='utf-8'), **self.dataset_defaults)

            chnk_grp = meta_grp.create_group(name='stored_chunks')
            chnk_grp.create_dataset(name='chrom', data=[], dtype=h5py.string_dtype(encoding='utf-8'), **self.dataset_defaults)
            chnk_grp.create_dataset(name='start', data=[], dtype='uint32', **self.dataset_defaults)
            chnk_grp.create_dataset(name='end', data=[], dtype='uint32', **self.dataset_defaults)

            # add chromosome group
            chrom_grp = expr_grp.create_group(f'chroms/')
            chrom_grp.create_dataset(name='name', data=list(trialn_header.chrom_lengths.keys()), dtype=h5py.string_dtype(encoding='utf-8'), **self.dataset_defaults)
            chrom_grp.create_dataset(name='length', data=list(trialn_header.chrom_lengths.values()), dtype='uint32', **self.dataset_defaults)

            # add a contacts/triplets group to store the cis-triplets
            cnct_grp = expr_grp.create_group(f'contacts/')
            trpl_grp = cnct_grp.create_group('triplets')
            trpl_grp.attrs['columns_order'] = [
                'flag',
                'start_A', 'start_B', 'start_C',
                'distance_AB', 'distance_AC', 'distance_BC',
                'count_A', 'count_B', 'count_C',
                'count_AB', 'count_AC', 'count_BC',
                'count_ABC',
            ]

            # iterate over each chrom, add anchors and triplets
            for chr_idx, (chrom_name, chrom_length) in enumerate(trialn_header.chrom_lengths.items()):
                if chrom_name == 'unmapped':
                    continue
                logger.debug(f'Initializing contact dataset: {chrom_name}')

                # define anchors
                anchor_edges = np.arange(0, chrom_length + anchor_width, anchor_width, dtype=int)
                ancr_grp = cnct_grp.create_group(f'anchors/{chrom_name}/')
                ancr_grp.create_dataset(name='start', data=anchor_edges[:-1], dtype='uint32', **self.dataset_defaults)
                ancr_grp.create_dataset(name='end', data=anchor_edges[1:], dtype='uint32', **self.dataset_defaults)

                # pairs group, to store 2D data
                # todo: to be added later, to support storage of 2D data

                # prepare cis-contact group
                cis_grp = trpl_grp.create_group(f'{chrom_name},{chrom_name},{chrom_name}/')
                n_anchor = len(anchor_edges) - 1
                n_anchor_per_slice = anchor_max_distance // anchor_width + 1
                n_pairs_per_slice = int((n_anchor_per_slice + 1) * n_anchor_per_slice / 2)
                n_triplets = n_anchor * n_pairs_per_slice

                # prepare the coordinates
                start_a, start_b, start_c = self.index2pos(range(n_triplets))

                # set flag, range=[0x0000 - 0xffff]
                # 0x0000: No flag is set
                # 0x0001: The Triplet coordinate is out of chromosome interval
                # 0x0002: At least one Triplet has an unassigned column/dataset
                is_out_of_bound = (start_a > chrom_length) | (start_b > chrom_length) | (start_c > chrom_length)
                flag = (
                        is_out_of_bound * 1
                        + np.ones(n_triplets) * 2
                ).astype('uint16')
                cis_grp.create_dataset(name='flag', shape=(n_triplets,), data=flag, dtype='uint16', **self.dataset_defaults)

                # triplets
                # todo: to be added later: To support sparse CSR storage
                # trp_grp.create_dataset(name='index_A', shape=(n_triplets, ), dtype='uint32', **self.dataset_defaults)
                # trp_grp.create_dataset(name='index_B', shape=(n_triplets, ), dtype='uint32', **self.dataset_defaults)
                # trp_grp.create_dataset(name='index_C', shape=(n_triplets, ), dtype='uint32', **self.dataset_defaults)
                cis_grp.create_dataset(name='start_A', shape=(n_triplets,), data=start_a, dtype='uint32', **self.dataset_defaults)
                cis_grp.create_dataset(name='start_B', shape=(n_triplets,), data=start_b, dtype='uint32', **self.dataset_defaults)
                cis_grp.create_dataset(name='start_C', shape=(n_triplets,), data=start_c, dtype='uint32', **self.dataset_defaults)
                cis_grp.create_dataset(name='distance_AB', shape=(n_triplets,), dtype='uint32', **self.dataset_defaults)
                cis_grp.create_dataset(name='distance_AC', shape=(n_triplets,), dtype='uint32', **self.dataset_defaults)
                cis_grp.create_dataset(name='distance_BC', shape=(n_triplets,), dtype='uint32', **self.dataset_defaults)
                cis_grp.create_dataset(name='count_A', shape=(n_triplets,), dtype='uint32', **self.dataset_defaults)
                cis_grp.create_dataset(name='count_B', shape=(n_triplets,), dtype='uint32', **self.dataset_defaults)
                cis_grp.create_dataset(name='count_C', shape=(n_triplets,), dtype='uint32', **self.dataset_defaults)
                cis_grp.create_dataset(name='count_AB', shape=(n_triplets,), dtype='uint32', **self.dataset_defaults)
                cis_grp.create_dataset(name='count_AC', shape=(n_triplets,), dtype='uint32', **self.dataset_defaults)
                cis_grp.create_dataset(name='count_BC', shape=(n_triplets,), dtype='uint32', **self.dataset_defaults)
                cis_grp.create_dataset(name='count_ABC', shape=(n_triplets,), dtype='uint32', **self.dataset_defaults)
                del start_a, start_b, start_c

    def pos2index(self, pos_a, pos_b, pos_c):
        if not isinstance(pos_a, np.ndarray):
            pos_a = np.array(pos_a)
        if not isinstance(pos_b, np.ndarray):
            pos_b = np.array(pos_b)
        if not isinstance(pos_c, np.ndarray):
            pos_c = np.array(pos_c)
        pos_a = pos_a.astype(int)
        pos_b = pos_b.astype(int)
        pos_c = pos_c.astype(int)
        assert np.all(pos_a <= pos_b)
        assert np.all(pos_b <= pos_c)

        dist_ab = pos_b - pos_a
        dist_ac = pos_c - pos_a
        dist_bc = pos_c - pos_b
        assert np.all((dist_ab >= 0) & (dist_ab <= self.properties['anchor_max_distance']))
        assert np.all((dist_ac >= 0) & (dist_ac <= self.properties['anchor_max_distance']))
        assert np.all((dist_bc >= 0) & (dist_bc <= self.properties['anchor_max_distance']))

        index_a = pos_a // self.properties['anchor_width']
        index_b = pos_b // self.properties['anchor_width'] - index_a
        index_c = pos_c // self.properties['anchor_width'] - index_a

        n_anchors_per_slice = self.properties['anchor_max_distance'] // self.properties['anchor_width'] + 1
        n_pairs_per_slice = int((n_anchors_per_slice + 1) * n_anchors_per_slice / 2)
        n_pairs_skipped = (index_b * (index_b + 1) / 2).astype(int)

        ds_row_index = index_a * n_pairs_per_slice + n_anchors_per_slice * index_b + index_c - n_pairs_skipped
        return ds_row_index

    def index2pos(self, ds_row_index):
        n_anchor_per_window = self.properties['anchor_max_distance'] // self.properties['anchor_width'] + 1
        n_pairs_per_window = int((n_anchor_per_window + 1) * n_anchor_per_window / 2)

        index_a, index_window = np.divmod(ds_row_index, n_pairs_per_window)
        index_b = n_anchor_per_window - 1 - np.floor((-1 + np.sqrt((2 * n_anchor_per_window + 1) ** 2 - 8 * (index_window + 1))) / 2).astype(int)
        index_c = index_window - index_b * (2 * n_anchor_per_window - index_b - 1) // 2

        return (
            index_a * self.properties['anchor_width'],
            index_a * self.properties['anchor_width'] + index_b * self.properties['anchor_width'],
            index_a * self.properties['anchor_width'] + index_c * self.properties['anchor_width'],
        )

    def create_columns(self, triplets_path, col_names, dtype='float32', exclusive=True):
        with HDF5Container(self.container_path, mode='a', exclusive=exclusive) as container:
            self.fetch_properties(exclusive=False)
            columns_order = list(self.properties['columns_order'])
            trpl_grp = container.h5_file[triplets_path]
            cis_chrom_names = list(trpl_grp.keys())

            # add each column, if not present
            h5_modified = False
            for col_name in col_names:
                if col_name not in columns_order:
                    logger.debug(f'Adding "{col_name}" column to Triplet\'s `column_order`')
                    columns_order.append(col_name)
                    h5_modified = True

                    # add the column to each cis_contact
                    for cis_chrom_name in cis_chrom_names:
                        cnct_grp = trpl_grp[cis_chrom_name]

                        # determine number of rows
                        stored_columns = list(cnct_grp.keys())
                        n_row = cnct_grp[stored_columns[0]].shape[0]

                        # adding dataset
                        logger.debug(f'Adding "{col_name}" dataset to: {cnct_grp.name}')
                        if col_name in cnct_grp:
                            logger.warning(f'"{col_name}" already exists in "{cnct_grp}"')
                            continue
                        cnct_grp.create_dataset(name=col_name, shape=(n_row,), dtype=dtype, **self.dataset_defaults)

            # update the related variables
            if h5_modified:

                # set the flag column to "unassigned"
                for cis_chrom_name in cis_chrom_names:
                    logger.debug(f'Setting "flag" column of "{cis_chrom_name}" triplets as: Unassigned')
                    trpl_grp[cis_chrom_name]['flag'][:] = np.bitwise_or(trpl_grp[cis_chrom_name]['flag'], 0x0002)

                # update meta-data
                logger.debug(f'Updating column-order as: {columns_order}')
                self.properties['columns_order'] = columns_order
                self.properties['stored_chunks'] = []
                self.store_properties(exclusive=False)

    def drop_columns(self, columns, exclusive=True):
        assert isinstance(columns, list), f'Please provide a list of column names. Currently given: {columns}'

        with HDF5Container(self.container_path, mode='a', exclusive=exclusive) as container:
            self.fetch_properties(exclusive=False)
            columns_order = list(self.properties['columns_order'])
            experiment_name = self.properties['experiment_name']
            cis_grp = container.h5_file['experiments'][experiment_name]['contacts']['triplets']

            # iterate over each given column
            for col_name in columns:
                for cis_chrom in cis_grp.keys():
                    trpt_grp = cis_grp[cis_chrom]
                    if col_name not in trpt_grp:
                        logger.warning(f'Could not find "{col_name}" column in: {cis_chrom}')
                    else:
                        logger.debug(f'Removing "{col_name}" column from: {cis_chrom}')
                        del trpt_grp[col_name]
                if col_name in columns_order:
                    logger.debug(f'Deleting "{col_name}" item from `columns_order` metadata')
                    columns_order.remove(col_name)

            # updating the meta-data
            logger.debug(f'Updating column-order as: {columns_order}')
            self.properties['columns_order'] = columns_order
            self.store_properties(exclusive=False)

    def store(
            self, triplets,
            chrom_a, chrom_b=None, chrom_c=None,
            start_a=None, start_b=None, start_c=None,
            exclusive=True, verify_coords=False
    ):

        # initialization
        if chrom_b is None:
            chrom_b = chrom_a
        if chrom_c is None:
            chrom_c = chrom_b
        if start_a is None:  # get coordinates from the Triplet container itself
            assert start_b is None
            assert start_c is None
            start_a = triplets['start_A']
            start_b = triplets['start_B']
            start_c = triplets['start_C']

        # store the values in columns
        with HDF5Container(self.container_path, mode='a', exclusive=exclusive) as container:

            # add columns if needed
            triplets_path = self.experiment_path + f'/contacts/triplets/'
            self.create_columns(triplets_path=triplets_path, col_names=list(triplets.keys()), exclusive=False)

            # storing the data
            row_indices = self.pos2index(start_a, start_b, start_c)
            cis_path = self.experiment_path + f'/contacts/triplets/{chrom_a},{chrom_b},{chrom_c}/'
            for col_name, col_values in triplets.items():
                if verify_coords and col_name.startswith('start_'):
                    assert np.array_equal(
                        container.retrieve(dataset_path=cis_path + col_name, row_indices=row_indices, exclusive=False),
                        col_values,
                    )
                else:
                    container.store(dataset_path=cis_path + col_name, row_indices=row_indices, values=col_values, exclusive=False)

            # set valid flag for stored triplets
            if 'flag' in container.h5_file[cis_path]:
                logger.debug(f'Setting the "flag" column of {len(row_indices):,d} stored triplets to: "Assigned"')
                flags = container.retrieve(dataset_path=f'{cis_path}/flag', row_indices=row_indices, exclusive=False)
                flags = np.bitwise_and(flags, 0xffff - 0x0002)
                container.store(dataset_path=f'{cis_path}/flag', row_indices=row_indices, values=flags, exclusive=False)
                del flags

    def fetch(
            self,
            chrom_a, start_a, end_a,
            chrom_b=None, start_b=None, end_b=None,
            chrom_c=None, start_c=None, end_c=None,
            exclusive=True, columns=None,
    ):

        # sanity checks
        if chrom_b is None:
            chrom_b = chrom_a
        if chrom_c is None:
            chrom_c = chrom_b
        if start_b is None:
            start_b = start_a
        if start_c is None:
            start_c = start_b
        if end_b is None:
            end_b = end_a + self.properties['anchor_max_distance']
        if end_c is None:
            end_c = end_a + self.properties['anchor_max_distance']
        assert isinstance(chrom_a, str) and len(chrom_a) > 0
        assert isinstance(chrom_b, str) and len(chrom_b) > 0
        assert isinstance(chrom_c, str) and len(chrom_c) > 0

        # determine row_indices
        start_idx = self.pos2index(start_a, start_b, start_c)
        end_idx = self.pos2index(end_a, end_b, end_c)
        assert start_idx <= end_idx
        row_indices = np.arange(start_idx, end_idx + 1)

        # prepare the container loader
        with HDF5Container(self.container_path, mode='r', exclusive=exclusive) as container:
            cis_path = f'{self.experiment_path}/contacts/triplets/{chrom_a},{chrom_b},{chrom_c}'
            if columns is None:
                columns = self.properties['columns_order']

            # iteratively load datasets
            triplets = Triplets()
            for col_name in columns:
                triplets[col_name] = container.retrieve(dataset_path=f'{cis_path}/{col_name}', row_indices=row_indices, exclusive=False)

            # get "flag" columns
            if 'flag' in triplets:
                flags = triplets['flag']
            else:
                flags = container.retrieve(dataset_path=f'{cis_path}/flag', row_indices=row_indices, exclusive=False)

        # filter the collected triplets, to overlap with the chromosome interval and the requested range
        within_chrom = np.bitwise_and(flags, 0x0001) != 0x0001
        within_interval = (
            (overlap(query_start=start_a, query_end=end_a, ref_start=triplets['start_A'], ref_end=triplets['start_A'] + self.properties['anchor_width'])) &
            (overlap(query_start=start_b, query_end=end_b, ref_start=triplets['start_B'], ref_end=triplets['start_B'] + self.properties['anchor_width'])) &
            (overlap(query_start=start_c, query_end=end_c, ref_start=triplets['start_C'], ref_end=triplets['start_C'] + self.properties['anchor_width']))
        )
        is_valid = within_chrom & within_interval
        if not is_valid.all():
            triplets = triplets[is_valid]
        del flags, within_chrom, within_interval, is_valid

        # prepare the return data
        return triplets

    def store_properties(self, exclusive=True):
        with HDF5Container(self.container_path, mode='a', exclusive=exclusive) as container:
            expr_grp = container.h5_file[self.experiment_path]
            for prop_name, prop_value in self.properties.items():
                if prop_name == 'stored_chunks':
                    if len(self.properties['stored_chunks']) == 0:
                        chunks_chrom, chunks_start, chunks_end = [], [], []
                    else:
                        chunks_chrom, chunks_start, chunks_end = zip(*self.properties['stored_chunks'])

                    del expr_grp[f'metadata/stored_chunks']
                    chk_grp = expr_grp.create_group(name='metadata/stored_chunks/')
                    chk_grp.create_dataset(name='chrom', data=chunks_chrom, dtype=h5py.string_dtype(encoding='utf-8'), **self.dataset_defaults)
                    chk_grp.create_dataset(name='start', data=chunks_start, dtype='uint32', **self.dataset_defaults)
                    chk_grp.create_dataset(name='end', data=chunks_end, dtype='uint32', **self.dataset_defaults)
                elif prop_name == 'chrom_names':
                    del expr_grp['chroms/name']
                    expr_grp['chroms'].create_dataset(name='name', data=prop_value, dtype=h5py.string_dtype(encoding='utf-8'), **self.dataset_defaults)
                elif prop_name == 'chrom_lengths':
                    del expr_grp['chroms/length']
                    expr_grp['chroms'].create_dataset(name='length', data=prop_value, dtype='uint32', **self.dataset_defaults)
                elif prop_name in ['sam_headers', 'triplix_headers']:
                    del expr_grp[f'metadata/{prop_name}']
                    expr_grp['metadata'].create_dataset(name=prop_name, data=prop_value, dtype=h5py.string_dtype(encoding='utf-8'), **self.dataset_defaults)
                elif prop_name == 'columns_order':
                    expr_grp['contacts/triplets'].attrs['columns_order'] = prop_value
                else:
                    expr_grp.attrs[prop_name] = prop_value

    def fetch_properties(self, exclusive=True):
        with HDF5Container(self.container_path, mode='r', exclusive=exclusive) as container:
            experiments = list(container.h5_file[f'/experiments'].keys())
            if len(experiments) > 1:
                raise NotImplementedError('Only single-experiment Triplet files are supported')
            experiment_name = experiments[0]
            expr_grp = container.h5_file[f'/experiments/{experiment_name}']

            # get experiment's metadata
            self.properties['sam_headers'] = [header.decode() for header in expr_grp['metadata/sam_headers'][()]]
            self.properties['triplix_headers'] = [header.decode() for header in expr_grp['metadata/triplix_headers'][()]]

            # collect stored chunks
            chnk_grp = expr_grp['metadata/stored_chunks/']
            chunks_chrom = [chr_name.decode() for chr_name in chnk_grp['chrom']]
            chunks_start = chnk_grp['start'][()]
            chunks_end = chnk_grp['end'][()]
            self.properties['stored_chunks'] = [[chrom, start, end] for chrom, start, end in zip(chunks_chrom, chunks_start, chunks_end)]

            # get experiment's attributes
            for attr_name in expr_grp.attrs.keys():
                self.properties[attr_name] = expr_grp.attrs[attr_name]

            # get columns order
            self.properties['columns_order'] = list(expr_grp['contacts/triplets'].attrs['columns_order'])

            # further processing of attributes
            self.properties['chrom_names'] = [chr_name.decode() for chr_name in expr_grp['chroms/name']]
            self.properties['chrom_lengths'] = list(expr_grp['chroms/length'])

    def store_attribute(self, dataset_path: str, name: str, value: Any, exclusive=True):
        with HDF5Container(self.container_path, mode='a', exclusive=exclusive) as container:
            ds = container.h5_file[dataset_path]
            ds.attrs[name] = value

    def add_stored_chunk(self, chunk, exclusive=True):
        if exclusive:
            self.lock()
        logger.debug(f'Adding to stored chunks: {chunk}')

        self.fetch_properties(exclusive=False)
        stored_chunks = list(self.properties['stored_chunks'])
        stored_chunks.append(chunk)
        chunks_merged = merge_intervals(
            intervals=stored_chunks,
            chroms_order=self.properties['chrom_names']
        )
        self.properties['stored_chunks'] = chunks_merged
        self.store_properties(exclusive=False)

        if exclusive:
            self.unlock()




import gzip
import pathlib
# from itertools import compress
import logging
import re
from typing import Union, Optional, Iterable
from collections import deque

import numpy as np
import pandas as pd
import pypairix
from triplix.core.header import TriplixHeader
from triplix.core import configurations
from triplix.core.utilities import overlap

COMMAND_NAME = 'triplix.tri-alignments'
logger = logging.getLogger(COMMAND_NAME)


class TriAlignments(dict):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.options = {
            'display.min_rows': 5
        }
        self.columns_dtypes = {
            'start_A': int, 'start_B': int, 'start_C': int,
            'end_A': int, 'end_B': int, 'end_C': int,
            'mapq_A': int, 'mapq_B': int, 'mapq_C': int,
        }
        # self.columns_integer = {col for col in self.return_columns if tri_container.columns_dtypes.get(col, None) is int}

        self.initialize()

    def initialize(self):
        for anc_lbl in ['A', 'B', 'C']:
            for anc_type in ['chrom', 'start', 'end']:
                self[f'{anc_type}_{anc_lbl}'] = []

    def __repr__(self):
        if 'start_A' not in self:
            return '<Miss-formatted Tri-alignments>: Missing `start_A` column'

        n_row = len(self['start_A'])
        n_show = self.options['display.min_rows']
        if n_row == 0 or n_row <= n_show:
            show_df = pd.DataFrame(self).astype(str)
        else:
            show_idxs = list(range(0, n_show + 1)) + list(range(n_row - n_show, n_row))
            show_df = pd.DataFrame({key: np.array(values)[show_idxs] for (key, values) in self.items()}, index=show_idxs)
            show_df = show_df.astype(str)
            show_df.loc[n_show, :] = '***'
            show_df.index = show_df.index.astype(str)
            show_df.rename({str(n_show): '***'}, axis=0, inplace=True)
        return '<Tri-alignments>:\n' + repr(show_df)

    def __len__(self):
        for column in self.keys():
            return len(self[column])

    def convert_dtypes(self):
        for hdr in self.keys():
            known_dtype = self.columns_dtypes.get(hdr, None)
            if known_dtype is int:
                self[hdr] = np.array(self[hdr], dtype=int)
            else:
                self[hdr] = np.array(self[hdr], dtype=object)

    def to_dataframe(self, *args, **kwargs):
        return pd.DataFrame(self, *args, **kwargs)


class TriAlignmentsContainer:

    def __init__(self, trialn_path: Union[str, pathlib.Path]):
        self.trialn_path = pathlib.Path(trialn_path).expanduser()
        self.header = TriplixHeader(file_path=trialn_path)
        self.column_names = self.header.column_names
        self.col2idx = {col: idx for idx, col in enumerate(self.column_names)}

        # initialize PyPairix object, if possible
        try:
            self.px_obj = pypairix.open(str(self.trialn_path))
            logger.debug(f'Loaded Pairix index of: {self.trialn_path}')
        except pypairix.PairixError:
            self.px_obj = None
            logger.warning(f'Failed to load Pairix index: {self.trialn_path}')

    def fetch(
            self,
            chrom_a: Optional[str] = None, start_a: Optional[float] = None, end_a: Optional[float] = None,
            chrom_b: Optional[str] = None, start_b: Optional[float] = None, end_b: Optional[float] = None,
            chrom_c: Optional[str] = None, start_c: Optional[float] = None, end_c: Optional[float] = None,
            batch_size=int(1e6), filter_c=True,
            mapping_quality: Optional[int] = None,
            return_columns: Optional[Iterable] = None,
            read_regex: str = None,
            experiment_regex: str = None,
    ):

        # set the default values
        if chrom_a is None:
            return self.gzip_iter(mapping_quality)
        else:
            if self.px_obj is None:
                index_filepath = self.trialn_path.parent / (self.trialn_path.name + '.px2')
                if not index_filepath.is_file():
                    raise FileNotFoundError(
                        f'Tri-alignment\'s index file ({index_filepath}) is not found. '
                        f'Please index the Tri-alignments file first using `triplix index` command.'
                    )
                raise pypairix.PairixError(f'Failed to load Tri-alignment\'s index file ({index_filepath}). ')

            if chrom_b is None:
                chrom_b = chrom_a
            if chrom_c is None:
                chrom_c = chrom_b
            if start_b is None:
                start_b = start_a
            if start_c is None:
                start_c = start_b
            if end_b is None:
                end_b = end_a
            if end_c is None:
                end_c = end_b
            assert isinstance(chrom_a, str) and len(chrom_a) > 0
            assert isinstance(chrom_b, str) and len(chrom_b) > 0
            assert isinstance(chrom_c, str) and len(chrom_c) > 0
            # assert np.isfinite(start_A)
            # assert np.isfinite(start_B)
            # assert np.isfinite(start_C)
            # assert np.isfinite(end_A)
            # assert np.isfinite(end_B)
            # assert np.isfinite(end_C)

            return TriAlignmentIterator(
                self,
                chrom_a=chrom_a, start_a=int(start_a), end_a=int(end_a),
                chrom_b=chrom_b, start_b=int(start_b), end_b=int(end_b),
                chrom_c=chrom_c, start_c=int(start_c), end_c=int(end_c),
                batch_size=batch_size, filter_c=filter_c, return_columns=return_columns,
                mapping_quality=mapping_quality,
                read_regex=read_regex,
                experiment_regex=experiment_regex,
            )

    def fetch_slice(
            self,
            vp_chrom, vp_start, vp_end,
            anchor_edges,
            anchor_max_distance=int(2e6),
            mapping_quality=5,

    ):
        n_anchor = len(anchor_edges) - 1
        counts = np.zeros([n_anchor, n_anchor], dtype=int)
        for aj in range(n_anchor):
            if anchor_edges[aj] < vp_start:
                continue
            if anchor_edges[aj] - vp_start > anchor_max_distance:
                break

            tri_alignments_iter = self.fetch(
                chrom_a=vp_chrom, start_a=vp_start, end_a=vp_end,
                chrom_b=vp_chrom, start_b=anchor_edges[aj], end_b=anchor_edges[aj + 1] - 1,
                chrom_c=vp_chrom, start_c=anchor_edges[aj], end_c=anchor_edges[-1] - 1,
                batch_size=int(1e6), mapping_quality=mapping_quality, return_columns=[
                    'read_name',
                    'chrom_C', 'start_C', 'end_C',
                    'mapq_A', 'mapq_B', 'mapq_C',
                ],
            )
            for tri_alignments in tri_alignments_iter:
                # tri_alignments.to_dataframe()

                # filter alignments with low mapping quality
                # is_mapped = (
                #         (tri_alignments['mapq_A'] >= mapping_quality) &
                #         (tri_alignments['mapq_B'] >= mapping_quality) &
                #         (tri_alignments['mapq_C'] >= mapping_quality)
                #
                # )
                # if not np.all(is_mapped):
                #     for hdr in tri_alignments.keys():
                #         tri_alignments[hdr] = tri_alignments[hdr][is_mapped]

                # find overlaps across bins
                start_idxs = np.searchsorted(anchor_edges, tri_alignments['start_C'], side='right') - 1
                end_idxs = np.searchsorted(anchor_edges, tri_alignments['end_C'], side='right') - 1

                # make sure each read contributes only once to each Triplet
                seen_ids = [set() for _ in range(n_anchor)]
                for start_idx, end_idx, read_id in zip(start_idxs, end_idxs, tri_alignments['read_name']):
                    for ak in range(start_idx, end_idx + 1):
                        if anchor_edges[ak] - vp_start > anchor_max_distance:
                            continue
                        if not aj <= ak < n_anchor:  # n_bin is needed, as the `alignment_end` could still fall outside `load_end`
                            continue
                        if read_id in seen_ids[ak]:
                            continue

                        seen_ids[ak].add(read_id)
                        counts[aj, ak] += 1
        return counts

    def gzip_iter(self, mapping_quality=None):
        with gzip.open(self.trialn_path, 'rt') as gz_file:

            # skipping the header lines
            for line in gz_file:
                if line.startswith('#column names:'):
                    break

            # iterate over the rest of the lines, returning a list of columns
            if mapping_quality is None:
                for line in gz_file:
                    yield line.rstrip().split('\t')
            else:
                for line in gz_file:
                    # filter for mapping quality
                    if (
                            int(line[self.col2idx['mapq_A']]) < mapping_quality
                            or int(line[self.col2idx['mapq_B']]) < mapping_quality
                            or int(line[self.col2idx['mapq_C']]) < mapping_quality
                    ):
                        continue
                    yield line.rstrip().split('\t')


class TriAlignmentIterator:
    """
    Collect Tri-alignments overlapping with a given a given tri-coordinates
    :param chrom_a: chromosome A of the Tri-alignments of interest
    :param chrom_b: chromosome B of the Tri-alignments of interest
    :param chrom_c: chromosome C of the Tri-alignments of interest
    :param start_a: start position of the A interval
    :param start_b: start position of the B interval
    :param start_c: start position of the C interval
    :param end_a: end position of the A interval
    :param end_b: end position of the B interval
    :param end_c: end position of the C interval
    :param batch_size: number of Tri-alignments to load in each iteration
    :param filter_c: whether the Tri-alignments should be filtered to have overlap with *_C interval
    :param mapping_quality: the minimum mapping quality that the alignments (of a tri-alignment) should have to be considered valid.

    Note:
        [start|end]_[A|B|C] : coordinates are inclusive on both sides
    """

    def __init__(
            self, tri_container: TriAlignmentsContainer,
            chrom_a: str, start_a: int, end_a: int,
            chrom_b: str, start_b: int, end_b: int,
            chrom_c: str, start_c: int, end_c: int,
            batch_size: int,
            filter_c=True, return_columns=None,
            mapping_quality=None,
            read_regex=None,
            experiment_regex=None,
    ):
        self.chrom_c = chrom_c
        self.start_c = start_c
        self.end_c = end_c
        self.batch_size = batch_size
        self.filter_C = filter_c
        self.mapping_quality = mapping_quality
        if read_regex is None:
            self.read_regex = None
        else:
            self.read_regex = re.compile(self.read_regex)
        if experiment_regex is None:
            self.experiment_regex = None
        else:
            self.experiment_regex = re.compile(experiment_regex)
        self.col2idx = tri_container.col2idx
        if return_columns is None:
            self.return_columns = tri_container.column_names
        else:
            self.return_columns = return_columns
        self.EOF = False

        if self.filter_C:
            headers_c = ['chrom_C', 'start_C', 'end_C']
            assert all(hdr in self.return_columns for hdr in headers_c), f'The following columns must be included in the `return_columns`: {headers_c}'

        # initialize query2D object
        self.trialn_iter = tri_container.px_obj.query2D(chrom_a, start_a, end_a,
                                                        chrom_b, start_b, end_b)

    def __iter__(self):
        return self

    def __next__(self):
        if self.EOF:
            raise StopIteration()

        # initialize a Tri-alignments container
        tri_alignments = TriAlignments()
        for col in self.return_columns:
            tri_alignments[col] = []

        # load a batch of tri-alignments
        n_load = 0
        for line in self.trialn_iter:

            # tri-alignment filters
            if self.mapping_quality is not None:
                is_mapped = (
                        int(line[self.col2idx['mapq_A']]) >= self.mapping_quality and
                        int(line[self.col2idx['mapq_B']]) >= self.mapping_quality and
                        int(line[self.col2idx['mapq_C']]) >= self.mapping_quality
                )
                if not is_mapped:
                    continue
            if self.read_regex is not None:
                read_name = line[self.col2idx['read_name']]
                if not self.read_regex.match(read_name):
                    continue
            if self.experiment_regex is not None:
                experiment_name = line[self.col2idx['experiment_name']]
                if not self.experiment_regex.match(experiment_name):
                    continue

            # adding the tri-alignment
            for hdr in tri_alignments.keys():
                tri_alignments[hdr].append(line[self.col2idx[hdr]])
            n_load += 1
            if n_load >= self.batch_size:
                break
        else:
            self.EOF = True

        # convert the columns dtypes
        tri_alignments.convert_dtypes()

        if n_load == 0:
            raise StopIteration()

        # filter rows, if needed
        if self.filter_C:
            has_ovl = (
                    (self.chrom_c == tri_alignments['chrom_C'])
                    & (self.start_c <= tri_alignments['end_C'])
                    & (self.end_c >= tri_alignments['start_C'])
            )
            if not np.all(has_ovl):
                for hdr in tri_alignments.keys():
                    tri_alignments[hdr] = tri_alignments[hdr][has_ovl]

        return tri_alignments



import copy
import pathlib
from dataclasses import dataclass

import pysam


@dataclass
class Alignment:
    chrom: str = None
    start: int = -1
    end: int = -1
    strand: str = None
    map_quality: int = -1
    flag: int = -1
    seq_start: int = -1
    seq_end: int = -1
    read_idx: int = -1
    read_name: str = None
    read_length: int = -1
    # cigar: str = None
    pysam_alignment: pysam.AlignedSegment = None


class BAMReader:
    def __init__(self, file_path):
        path_obj = pathlib.Path(file_path).expanduser()
        assert path_obj.is_file(), f'File not found: {file_path}'
        self.file_path = file_path
        _verbosity_level = pysam.set_verbosity(0)
        self.file_handle = pysam.AlignmentFile(path_obj, 'r')
        pysam.set_verbosity(_verbosity_level)
        self.chroms = [chrom['SN'] for chrom in self.file_handle.header['SQ']]
        self.chrom_sizes = [chrom['LN'] for chrom in self.file_handle.header['SQ']]

    def get_reads(self, return_pysam_object=False):
        read_idx = 0
        alignments = []
        for aln_idx, aln in enumerate(self.file_handle):

            alignment = Alignment(
                chrom=aln.reference_name,
                start=aln.reference_start,
                end=aln.reference_end,
                strand='-' if aln.is_reverse else '+',
                map_quality=aln.mapping_quality,
                flag=aln.flag,
                read_idx=read_idx,
                read_name=aln.query_name,
                read_length=aln.infer_read_length(),
                # cigar=aln.cigarstring,
            )

            # infer the alignment position over read
            # PySam regular functions wont work: https://github.com/pysam-developers/pysam/issues/903#issuecomment-1240695780
            # S	BAM_CSOFT_CLIP	4
            # H	BAM_CHARD_CLIP	5
            skip_start = aln.cigartuples[0][1] if aln.cigartuples[0][0] in [4, 5] else 0
            skip_end = aln.cigartuples[-1][1] if aln.cigartuples[-1][0] in [4, 5] else 0
            if aln.is_reverse:
                alignment.seq_start = skip_end
                alignment.seq_end = alignment.read_length - skip_start
            else:
                alignment.seq_start = skip_start
                alignment.seq_end = alignment.read_length - skip_end

            if return_pysam_object:
                alignment.pysam_alignment = copy.deepcopy(aln)

            # check if the current alignment should be appended
            if len(alignments) == 0 or alignments[0].read_name == alignment.read_name:
                alignments.append(alignment)
            else:
                yield alignments
                read_idx += 1
                alignment.read_idx = read_idx
                alignments = [alignment]
        yield alignments

    # def get_alignments(self, chrom=None, start=None, end=None, return_pysam_object=False):
    #     if chrom is None:
    #         alignments_iter = self.file_handle
    #     else:
    #         alignments_iter = self.file_handle.fetch(contig=chrom, start=start, end=end)




# todo: add an argument to ignore adding Triplets that are overlapping with each other within a read


import gzip
import logging
import pathlib
import copy
import time

import h5py
import numpy as np

from triplix.core import configurations
from triplix.core.header import TriplixHeader
from triplix.core.concatemers import ConcatemersContainer
from triplix.core.utilities import generate_prog_id

# produce pairix:
# en=sample_100e3_0;
# pairtools parse --chroms-path=${HOME}/bulk/my_works/delaat_repositories/hic_processing/references/hg38.chrom.sizes.tsv --output=./pairs/${en}.parsed.pairsam.gz --add-columns mapq bams/${en}.bam
# pairtools sort --output=./pairs/${en}.sorted.pairsam.gz ./pairs/${en}.parsed.pairsam.gz
# pairtools merge --output=./pairs/sample_100e3.merged.gz ./pairs/*.sorted.pairsam.gz

FORMAT_VERSION = '1.0.0'
COMMAND_NAME = 'triplix.tri-alignment'
COMMAND_ID = f'{COMMAND_NAME}>{generate_prog_id()}'
logger = logging.getLogger(COMMAND_NAME)


class TriAlignmentExtractor:
    def __init__(self, concatemers_path, output_dir=None, output_name=None):
        self.concatemers_path = pathlib.Path(concatemers_path).expanduser()
        self.header = TriplixHeader(file_path=concatemers_path)
        self.experiment_name = self.header.experiment_name

        if output_dir is None:
            output_dir = self.concatemers_path.parent
        if output_name is None:
            output_name = str(self.concatemers_path.name).replace('.concatemers.h5', '') + '.tri-alignments.tsv.bgz'
        self.output_path = pathlib.Path(output_dir).expanduser() / output_name
        self.output_path.parent.mkdir(parents=True, exist_ok=True)

    def extract_tri_alignments(self):

        # initialize output files and add Triplix headers
        output_files = {}
        for chrom in self.header.chrom_lengths.keys():
            if chrom == 'unmapped':
                continue
            output_path = str(self.output_path).replace('.tri-alignments.tsv.bgz', f'.{chrom}.tri-alignments.tsv.bgz')
            logger.debug(f'Initializing output file: {output_path}')
            output_files[chrom] = gzip.open(output_path, 'wt', compresslevel=3)

            # include a PG header SAM headers to describe the Triplet generation procedure
            taln_header = copy.deepcopy(self.header)
            taln_header.stored_chunks = [(chrom, 0, self.header.chrom_lengths[chrom])]
            taln_header.process_history.add_pg(dict(ID=COMMAND_ID, PN=COMMAND_NAME, VN=configurations.configs['version'], PS=[taln_header.process_history.destination_id]))
            output_files[chrom].write(taln_header.to_string() + '\n')

        # get ENUM data mappers
        cctm_container = ConcatemersContainer(file_path=self.concatemers_path)
        chrom2num = h5py.check_enum_dtype(cctm_container.h5_file['/concatemers/chrom_num'].dtype)
        strnd2num = h5py.check_enum_dtype(cctm_container.h5_file['/concatemers/strand'].dtype)
        num2chrom = {value: key for key, value in chrom2num.items()}
        num2strnd = {value: key for key, value in strnd2num.items()}

        # iterate over concatemers
        logger.info(f'Extracting cis tri-alignments from each concatemer ...')
        concatemers_iter = cctm_container.iter(return_name=True, return_length=True)
        n_concatemer = cctm_container.h5_file['/reads/name'].shape[0]
        n_tri_alignments = 0
        logger.log_time = time.time()
        for cnct_idx, concatemer in enumerate(concatemers_iter):
            # pd.DataFrame(concatemer)
            if time.time() - logger.log_time > configurations.configs['log_interval']:
                logger.log_time = time.time()
                logger.info(f'\t{cnct_idx:,d}/{n_concatemer:,d} concatemers are processed. '
                            f'{n_tri_alignments:,d} tri-alignments are generated so far ...')
            n_frag = len(concatemer['read_idx'])

            # identify multi-way interactions
            captured_chrs, captured_freq = np.unique(concatemer['chrom_num'], return_counts=True)
            captured_chrs = captured_chrs[captured_freq > 2]

            # iterate over chromosomes with multi-way captures
            for ci, chrom_cis in enumerate(captured_chrs):
                cis_idxs = np.where(concatemer['chrom_num'] == chrom_cis)[0]
                n_cis = len(cis_idxs)
                cis_idxs = cis_idxs[np.argsort(concatemer['start'][cis_idxs])]

                # iterate over cis alignments
                for cis_idx_i in range(n_cis):
                    for cis_idx_j in range(cis_idx_i + 1, n_cis):
                        for cis_idx_k in range(cis_idx_j + 1, n_cis):
                            i = cis_idxs[cis_idx_i]
                            j = cis_idxs[cis_idx_j]
                            k = cis_idxs[cis_idx_k]

                            tri_alignment_str = (
                                    f"{num2chrom[concatemer['chrom_num'][i]]}\t{concatemer['start'][i]}\t{concatemer['end'][i]}\t"
                                    + f"{num2chrom[concatemer['chrom_num'][j]]}\t{concatemer['start'][j]}\t{concatemer['end'][j]}\t"
                                    + f"{num2chrom[concatemer['chrom_num'][k]]}\t{concatemer['start'][k]}\t{concatemer['end'][k]}\t"
                                    + f"{num2strnd[concatemer['strand'][i]]:s}\t"
                                    + f"{num2strnd[concatemer['strand'][j]]:s}\t"
                                    + f"{num2strnd[concatemer['strand'][k]]:s}\t"
                                    + f"{concatemer['map_quality'][i]:d}\t"
                                    + f"{concatemer['map_quality'][j]:d}\t"
                                    + f"{concatemer['map_quality'][k]:d}\t"
                                    + f"{i:d}\t"
                                    + f"{j:d}\t"
                                    + f"{k:d}\t"
                                    + f"{concatemer['read_name']:s}\t"
                                    + f"{concatemer['read_length']:d}\t"
                                    + f"{n_frag:d}\t"
                                    + f"{self.experiment_name:s}\n"
                            )
                            output_files[num2chrom[chrom_cis]].write(tri_alignment_str)
                            n_tri_alignments += 1
        logger.info(f'All {n_tri_alignments:,d} tri-alignments are exported successfully.')

        # closing the output file handles
        logger.info(f'Closing the output file handles:')
        for chrom, file_handle in output_files.items():
            file_handle.close()
            logger.debug(f'File closed: {file_handle.name}')

